# flake8: noqa

import logging
import os
import sys
import atexit

logger = logging.getLogger(__name__)

# user code can override this null handler
logger.addHandler(logging.NullHandler())

def library_name(name, abi_number, libprefix='lib'):
    is_windows = os.name == 'nt'
    is_mac = sys.platform == 'darwin'

    if is_windows:
        return f'{libprefix}{name}-{abi_number}.dll'
    elif is_mac:
        return f'{libprefix}{name}.{abi_number}.dylib'
    else:
        return f'{libprefix}{name}.so.{abi_number}'

# pull in our module version number
from .version import __version__

# try to import our binary interface ... if that works, we are in API mode
API_mode = False
try:
    import _libvips

    logger.debug('Loaded binary module _libvips')

    ffi = _libvips.ffi
    vips_lib = _libvips.lib
    glib_lib = _libvips.lib
    gobject_lib = _libvips.lib

    # now check that the binary wrapper is for the same version of libvips that
    # we find ourseleves linking to at runtime ... if it isn't, we must fall
    # back to ABI mode
    lib_major = vips_lib.vips_version(0)
    lib_minor = vips_lib.vips_version(1)
    wrap_major = vips_lib.VIPS_MAJOR_VERSION
    wrap_minor = vips_lib.VIPS_MINOR_VERSION
    logger.debug(f'Module generated for libvips {wrap_major}.{wrap_minor}')
    logger.debug(f'Linked to libvips {lib_major}.{lib_minor}')

    if wrap_major != lib_major or wrap_minor != lib_minor:
        raise Exception('bad wrapper version')

    API_mode = True

except Exception as e:
    logger.debug(f'Binary module load failed: {e}')
    logger.debug('Falling back to ABI mode')

    from cffi import FFI

    ffi = FFI()

    vips_lib = ffi.dlopen(library_name('vips', 42))
    glib_lib = vips_lib
    gobject_lib = vips_lib

    logger.debug('Loaded lib %s', vips_lib)

    ffi.cdef('''
        int vips_init (const char* argv0);
        int vips_version (int flag);
    ''')

if vips_lib.vips_init(sys.argv[0].encode()) != 0:
    raise Exception('unable to init libvips')

logger.debug('Inited libvips')

if not API_mode:
    from .vdecls import cdefs

    major = vips_lib.vips_version(0)
    minor = vips_lib.vips_version(1)
    micro = vips_lib.vips_version(2)
    features = {
        'major': major,
        'minor': minor,
        'micro': micro,
        'api': False,
    }

    ffi.cdef(cdefs(features))

    # We can sometimes get dependent libraries from libvips -- either the platform
    # will open dependencies for us automatically, or the libvips binary has been
    # built to includes all main dependencies (common on windows, can happen
    # elsewhere).
    #
    # We must get glib functions from libvips if we can, since it will be the
    # one that libvips itself is using, and they will share runtime types.
    try:
        is_unified = gobject_lib.g_type_from_name(b'VipsImage') != 0
    except Exception:
        is_unified = False

    if not is_unified:
        libprefix = '' if os.name == 'nt' else 'lib'
        glib_lib = ffi.dlopen(library_name('glib-2.0', 0, libprefix))
        gobject_lib = ffi.dlopen(library_name('gobject-2.0', 0, libprefix))

        logger.debug('Loaded lib %s', glib_lib)
        logger.debug('Loaded lib %s', gobject_lib)


from .error import *

# redirect all vips warnings to logging

class GLogLevelFlags(object):
    # log flags
    FLAG_RECURSION          = 1 << 0
    FLAG_FATAL              = 1 << 1

    # GLib log levels
    LEVEL_ERROR             = 1 << 2       # always fatal 
    LEVEL_CRITICAL          = 1 << 3
    LEVEL_WARNING           = 1 << 4
    LEVEL_MESSAGE           = 1 << 5
    LEVEL_INFO              = 1 << 6
    LEVEL_DEBUG             = 1 << 7

    LEVEL_TO_LOGGER = {
        LEVEL_DEBUG: 10,
        LEVEL_INFO: 20,
        LEVEL_MESSAGE: 20,
        LEVEL_WARNING: 30,
        LEVEL_ERROR: 40,
        LEVEL_CRITICAL: 50,
    }


if API_mode:
    @ffi.def_extern()
    def _log_handler_callback(domain, level, message, user_data):
        logger.log(GLogLevelFlags.LEVEL_TO_LOGGER[level],
                   f'{_to_string(domain)}: {_to_string(message)}')

    # keep a ref to the cb to stop it being GCd
    _log_handler_cb = glib_lib._log_handler_callback
else:
    def _log_handler_callback(domain, level, message, user_data):
        logger.log(GLogLevelFlags.LEVEL_TO_LOGGER[level],
                   f'{_to_string(domain)}: {_to_string(message)}')

    # keep a ref to the cb to stop it being GCd
    _log_handler_cb = ffi.callback('GLogFunc', _log_handler_callback)

_log_handler_id = glib_lib.g_log_set_handler(_to_bytes('VIPS'),
                                             GLogLevelFlags.LEVEL_DEBUG |
                                             GLogLevelFlags.LEVEL_INFO |
                                             GLogLevelFlags.LEVEL_MESSAGE |
                                             GLogLevelFlags.LEVEL_WARNING |
                                             GLogLevelFlags.LEVEL_CRITICAL |
                                             GLogLevelFlags.LEVEL_ERROR |
                                             GLogLevelFlags.FLAG_FATAL |
                                             GLogLevelFlags.FLAG_RECURSION,
                                             _log_handler_cb, ffi.NULL)

# ffi doesn't like us looking up methods during shutdown: make a note of the
# remove handler here
_remove_handler = glib_lib.g_log_remove_handler

# we must remove the handler on exit or libvips may try to run the callback
# during shutdown
def _remove_log_handler():
    global _log_handler_id
    global _remove_handler

    if _log_handler_id:
        _remove_handler(_to_bytes('VIPS'), _log_handler_id)
        _log_handler_id = None


atexit.register(_remove_log_handler)

from .enums import *
from .base import *
from .gobject import *
from .gvalue import *
from .vobject import *
from .vinterpolate import *
from .vconnection import *
from .vsource import *
from .vsourcecustom import *
from .vtarget import *
from .vtargetcustom import *
from .voperation import *
from .vimage import *
from .vregion import *

__all__ = ['API_mode']



import logging

import pyvips
from pyvips import ffi, vips_lib, Error, _to_bytes

logger = logging.getLogger(__name__)


class Source(pyvips.Connection):
    """An input connection.

    """

    def __init__(self, pointer):
        # logger.debug('Operation.__init__: pointer = %s', pointer)
        super(Source, self).__init__(pointer)

    @staticmethod
    def new_from_descriptor(descriptor):
        """Make a new source from a file descriptor (a small integer).

        Make a new source that is attached to the descriptor. For example::

            source = pyvips.Source.new_from_descriptor(0)

        Makes a descriptor attached to stdin.

        You can pass this source to (for example) :meth:`new_from_source`.

        """

        # logger.debug('VipsSource.new_from_descriptor: descriptor = %d',
        #   descriptor)

        # sources are mutable, so we can't use the cache
        pointer = vips_lib.vips_source_new_from_descriptor(descriptor)
        if pointer == ffi.NULL:
            raise Error(f"can't create source from descriptor {descriptor}")

        return Source(pointer)

    @staticmethod
    def new_from_file(filename):
        """Make a new source from a filename.

        Make a new source that is attached to the named file. For example::

            source = pyvips.Source.new_from_file("myfile.jpg")

        You can pass this source to (for example) :meth:`new_from_source`.

        """

        # logger.debug('VipsSource.new_from_file: filename = %s',
        #              filename)

        pointer = vips_lib.vips_source_new_from_file(_to_bytes(filename))
        if pointer == ffi.NULL:
            raise Error(f"can't create source from filename {filename}")

        return Source(pointer)

    @staticmethod
    def new_from_memory(data):
        """Make a new source from a memory object.

        Make a new source that is attached to the memory object. For example::

            source = pyvips.Source.new_from_memory("myfile.jpg")

        You can pass this source to (for example) :meth:`new_from_source`.

        The memory object can be anything that supports the Python buffer or
        memoryview protocol.

        """

        # logger.debug('VipsSource.new_from_memory:')

        start = ffi.from_buffer(data)
        nbytes = data.nbytes if hasattr(data, 'nbytes') else len(data)

        pointer = vips_lib.vips_source_new_from_memory(start, nbytes)
        if pointer == ffi.NULL:
            raise Error("can't create input source from memory")

        source = Source(pointer)

        # keep a secret reference to the input data to make sure it's not GCed
        source._references = [data]

        return source


__all__ = ['Source']



# all the C decls for pyvips

# we keep these together to make switching between ABI and API modes simpler

# we have to pass in the libvips version, since it can come from either
# pkg-config in compile.py (in API mode) or libvips itself in __init__.py
# (in ABI mode)

import sys


def _at_least(features, x, y):
    return features['major'] > x or (features['major'] == x and
                                     features['minor'] >= y)


def cdefs(features):
    """Return the C API declarations for libvips.

    features is a dict with the features we want. Some features were only
    added in later libvips, for example, and some need to be disabled in
    some FFI modes.

    """

    # we need the glib names for these types
    code = '''
        typedef uint32_t guint32;
        typedef int32_t gint32;
        typedef uint64_t guint64;
        typedef int64_t gint64;
    '''

    # apparently the safest way to do this
    is_64bits = sys.maxsize > 2 ** 32

    # GType is an int the size of a pointer ... I don't think we can just use
    # size_t, sadly
    if is_64bits:
        code += '''
            typedef guint64 GType;
        '''
    else:
        code += '''
            typedef guint32 GType;
        '''

    # ... means opaque
    code += '''
        typedef void (*GLogFunc) (const char* log_domain,
            int log_level,
            const char* message, void* user_data);
        int g_log_set_handler (const char* log_domain,
            int log_levels,
            GLogFunc log_func, void* user_data);

        extern "Python" void _log_handler_callback (const char*, int,
            const char*, void*);

        void g_log_remove_handler (const char* log_domain, int handler_id);

        typedef ... VipsImage;

        void* g_malloc (size_t size);
        void g_free (void* data);

        void vips_leak_set (int leak);

        GType vips_type_find (const char* basename, const char* nickname);
        const char* vips_nickname_find (GType type);

        const char* g_type_name (GType gtype);
        GType g_type_from_name (const char* name);

        typedef void* (*VipsTypeMap2Fn) (GType type, void* a, void* b);
        void* vips_type_map (GType base, void* fn, void* a, void* b);

        void vips_shutdown (void);

        const char* vips_error_buffer (void);
        void vips_error_clear (void);
        void vips_error_freeze (void);
        void vips_error_thaw (void);

        typedef struct _GValue {
            GType g_type;
            union {
                guint64 v_uint64;

                // more
            } data[2];
        } GValue;

        void g_value_init (GValue* value, GType gtype);
        void g_value_unset (GValue* value);
        GType g_type_fundamental (GType gtype);

        int vips_enum_from_nick (const char* domain,
            GType gtype, const char* str);
        int vips_flags_from_nick (const char* domain,
            GType gtype, const char* nick);
        const char *vips_enum_nick (GType gtype, int value);

        void g_value_set_boolean (GValue* value, int v_boolean);
        void g_value_set_int (GValue* value, int i);
        void g_value_set_uint64 (GValue* value, guint64 ull);
        void g_value_set_double (GValue* value, double d);
        void g_value_set_enum (GValue* value, int e);
        void g_value_set_flags (GValue* value, unsigned int f);
        void g_value_set_string (GValue* value, const char* str);
        void vips_value_set_ref_string (GValue* value, const char* str);
        void g_value_set_object (GValue* value, void* object);
        void vips_value_set_array_double (GValue* value,
            const double* array, int n );
        void vips_value_set_array_int (GValue* value,
            const int* array, int n );
        void vips_value_set_array_image (GValue *value, int n);

        int g_value_get_boolean (const GValue* value);
        int g_value_get_int (GValue* value);
        guint64 g_value_get_uint64 (GValue* value);
        double g_value_get_double (GValue* value);
        int g_value_get_enum (GValue* value);
        unsigned int g_value_get_flags (GValue* value);
        const char* g_value_get_string (GValue* value);
        const char* vips_value_get_ref_string (const GValue* value,
            size_t* length);
        void* g_value_get_object (GValue* value);
        double* vips_value_get_array_double (const GValue* value, int* n);
        int* vips_value_get_array_int (const GValue* value, int* n);
        VipsImage** vips_value_get_array_image (const GValue* value, int* n);
        void* vips_value_get_blob (const GValue* value, size_t* length);

        // need to make some of these by hand
        GType vips_interpretation_get_type (void);
        GType vips_operation_flags_get_type (void);
        GType vips_band_format_get_type (void);

        typedef ... GData;

        typedef struct _GTypeClass {
            GType g_type;
        } GTypeClass;

        typedef struct _GTypeInstance {
            GTypeClass *g_class;
        } GTypeInstance;

        typedef struct _GObject {
            GTypeInstance g_type_instance;

            unsigned int ref_count;
            GData *qdata;
        } GObject;

        typedef struct _GParamSpec {
            GTypeInstance g_type_instance;

            const char* name;
            unsigned int flags;
            GType value_type;
            GType owner_type;

            // private, but cffi in API mode needs these to be able to get the
            // offset of any member
            char* _nick;
            char* _blurb;
            GData* qdata;
            unsigned int ref_count;
            unsigned int param_id;
        } GParamSpec;

        typedef struct _GEnumValue {
            int value;

            const char *value_name;
            const char *value_nick;
        } GEnumValue;

        typedef struct _GEnumClass {
            GTypeClass g_type_class;

            int minimum;
            int maximum;
            unsigned int n_values;
            GEnumValue *values;
        } GEnumClass;

        typedef struct _GFlagsValue {
            unsigned int value;

            const char *value_name;
            const char *value_nick;
        } GFlagsValue;

        typedef struct _GFlagsClass {
            GTypeClass g_type_class;

            unsigned int mask;
            unsigned int n_values;
            GFlagsValue *values;
        } GFlagsClass;

        void* g_type_class_ref (GType type);

        void* g_object_new (GType type, void*);
        void g_object_ref (void* object);
        void g_object_unref (void* object);

        void g_object_set_property (GObject* object,
            const char *name, GValue* value);
        void g_object_get_property (GObject* object,
            const char* name, GValue* value);

        void vips_image_invalidate_all (VipsImage* image);

        typedef void (*GCallback)(void);
        typedef void (*GClosureNotify)(void* data, struct _GClosure *);
        long g_signal_connect_data (GObject* object,
            const char* detailed_signal,
            GCallback c_handler,
            void* data,
            GClosureNotify destroy_data,
            int connect_flags);

        extern "Python" void _marshal_image_progress (VipsImage*,
            void*, void*);

        void vips_image_set_progress (VipsImage* image, int progress);
        void vips_image_set_kill (VipsImage* image, int kill);

        typedef ... GTimer;

        typedef struct _VipsProgress {
            VipsImage* im;

            int run;
            int eta;
            gint64 tpels;
            gint64 npels;
            int percent;
            GTimer* start;
        } VipsProgress;

        typedef ... VipsObject;

        typedef ... VipsObjectClass;

        typedef struct _VipsArgument {
            GParamSpec *pspec;
        } VipsArgument;

        typedef struct _VipsArgumentInstance {
            VipsArgument parent;

            // more
        } VipsArgumentInstance;

        typedef enum _VipsArgumentFlags {
            VIPS_ARGUMENT_NONE = 0,
            VIPS_ARGUMENT_REQUIRED = 1,
            VIPS_ARGUMENT_CONSTRUCT = 2,
            VIPS_ARGUMENT_SET_ONCE = 4,
            VIPS_ARGUMENT_SET_ALWAYS = 8,
            VIPS_ARGUMENT_INPUT = 16,
            VIPS_ARGUMENT_OUTPUT = 32,
            VIPS_ARGUMENT_DEPRECATED = 64,
            VIPS_ARGUMENT_MODIFY = 128
        } VipsArgumentFlags;

        typedef struct _VipsArgumentClass {
            VipsArgument parent;

            VipsObjectClass *object_class;
            VipsArgumentFlags flags;
            int priority;
            unsigned int offset;
        } VipsArgumentClass;

        int vips_object_get_argument (VipsObject* object,
            const char *name, GParamSpec** pspec,
            VipsArgumentClass** argument_class,
            VipsArgumentInstance** argument_instance);

        void vips_object_print_all (void);

        int vips_object_set_from_string (VipsObject* object,
            const char* options);

        const char* vips_object_get_description (VipsObject* object);

        const char* g_param_spec_get_blurb (GParamSpec* pspec);

        const char* vips_foreign_find_load (const char* name);
        const char* vips_foreign_find_load_buffer (const void* data,
            size_t size);
        const char* vips_foreign_find_save (const char* name);
        const char* vips_foreign_find_save_buffer (const char* suffix);

        VipsImage* vips_image_new_matrix_from_array (int width, int height,
                const double* array, int size);
        VipsImage* vips_image_new_from_memory (const void* data, size_t size,
                int width, int height, int bands, int format);

        VipsImage* vips_image_copy_memory (VipsImage* image);

        GType vips_image_get_typeof (const VipsImage* image,
            const char* name);
        int vips_image_get (const VipsImage* image,
            const char* name, GValue* value_copy);
        void vips_image_set (VipsImage* image,
            const char* name, GValue* value);
        int vips_image_remove (VipsImage* image, const char* name);

        char* vips_filename_get_filename (const char* vips_filename);
        char* vips_filename_get_options (const char* vips_filename);

        VipsImage* vips_image_new_temp_file (const char* format);

        int vips_image_write (VipsImage* image, VipsImage* out);
        void* vips_image_write_to_memory (VipsImage* in, size_t* size_out);

        typedef ... VipsInterpolate;

        VipsInterpolate* vips_interpolate_new (const char* name);

        typedef ... VipsOperation;

        VipsOperation* vips_operation_new (const char* name);

        typedef void* (*VipsArgumentMapFn) (VipsObject* object,
            GParamSpec* pspec,
            VipsArgumentClass* argument_class,
            VipsArgumentInstance* argument_instance,
            void* a, void* b);

        void* vips_argument_map (VipsObject* object,
            VipsArgumentMapFn fn, void* a, void* b);

        typedef ... VipsRegion;

        VipsRegion* vips_region_new (VipsImage*);

        VipsOperation* vips_cache_operation_build (VipsOperation* operation);
        void vips_object_unref_outputs (VipsObject* object);

        int vips_operation_get_flags (VipsOperation* operation);

        void vips_cache_set_max (int max);
        void vips_cache_set_max_mem (size_t max_mem);
        void vips_cache_set_max_files (int max_files);
        void vips_cache_set_trace (int trace);

        int vips_cache_get_max();
        int vips_cache_get_size();
        size_t vips_cache_get_max_mem();
        int vips_cache_get_max_files();

    '''

    # we must only define this in ABI mode ... in API mode we use
    # vips_value_set_blob_free in a backwards compatible way
    if not features['api']:
        code += '''
            typedef void (*FreeFn)(void* a);
            void vips_value_set_blob (GValue* value,
                FreeFn free_fn, void* data, size_t length);
        '''

    if _at_least(features, 8, 5):
        code += '''
            char** vips_image_get_fields (VipsImage* image);
            int vips_image_hasalpha (VipsImage* image);

        '''

    if _at_least(features, 8, 6):
        code += '''
            GType vips_blend_mode_get_type (void);
            void vips_value_set_blob_free (GValue* value,
                void* data, size_t length);

        '''

    if _at_least(features, 8, 7):
        code += '''
            int vips_object_get_args (VipsObject* object,
                const char*** names, int** flags, int* n_args);

        '''

    if _at_least(features, 8, 8):
        code += '''
            char** vips_foreign_get_suffixes (void);

            void* vips_region_fetch (VipsRegion*, int, int, int, int,
                size_t* length);
            int vips_region_width (VipsRegion*);
            int vips_region_height (VipsRegion*);
            int vips_image_get_page_height (VipsImage*);
            int vips_image_get_n_pages (VipsImage*);

        '''

    if _at_least(features, 8, 9):
        code += '''
            typedef ... VipsConnection;

            const char* vips_connection_filename (VipsConnection* stream);
            const char* vips_connection_nick (VipsConnection* stream);

            typedef ... VipsSource;

            VipsSource* vips_source_new_from_descriptor (int descriptor);
            VipsSource* vips_source_new_from_file (const char* filename);
            VipsSource* vips_source_new_from_memory (const void* data,
                size_t size);

            typedef ... VipsSourceCustom;

            VipsSourceCustom* vips_source_custom_new (void);

            extern "Python" gint64 _marshal_read (VipsSource*,
                void*, gint64, void*);
            extern "Python" gint64 _marshal_seek (VipsSource*,
                gint64, int, void*);

            typedef ... VipsTarget;

            VipsTarget* vips_target_new_to_descriptor (int descriptor);
            VipsTarget* vips_target_new_to_file (const char* filename);
            VipsTarget* vips_target_new_to_memory (void);

            typedef ... VipsTargetCustom;

            VipsTargetCustom* vips_target_custom_new (void);

            extern "Python" gint64 _marshal_write (VipsTarget*,
                void*, gint64, void*);
            extern "Python" void _marshal_finish (VipsTarget*,
                void*);

            const char* vips_foreign_find_load_source (VipsSource *source);
            const char* vips_foreign_find_save_target (const char* suffix);

        '''

    if _at_least(features, 8, 13):
        code += '''
            extern "Python" int _marshal_end (VipsTarget*,
                void*);

            void vips_block_untrusted_set (int state);
            void vips_operation_block_set (const char *name, int state);

        '''

    # we must only define these in API mode ... in ABI mode we need to call
    # these things earlier
    if features['api']:
        code += '''
            int vips_init (const char* argv0);
            int vips_version (int flag);
        '''

    # ... means inherit from C defines
    code += '''
        #define VIPS_MAJOR_VERSION ...
        #define VIPS_MINOR_VERSION ...
        #define VIPS_MICRO_VERSION ...
    '''

    # add contents of features as a comment ... handy for debugging
    for key, value in features.items():
        code += f'//{key} = {value}\n'

    return code


__all__ = ['cdefs']



import logging

import pyvips
from pyvips import ffi, vips_lib, Error, _to_bytes

logger = logging.getLogger(__name__)


class Target(pyvips.Connection):
    """An output connection.

    """

    def __init__(self, pointer):
        # logger.debug('Operation.__init__: pointer = %s', pointer)
        super(Target, self).__init__(pointer)

    @staticmethod
    def new_to_descriptor(descriptor):
        """Make a new target to write to a file descriptor (a small
        integer).

        Make a new target that is attached to the descriptor. For example::

            target = pyvips.Target.new_to_descriptor(1)

        Makes a descriptor attached to stdout.

        You can pass this target to (for example) :meth:`write_to_target`.

        """

        # logger.debug('VipsTarget.new_to_descriptor: descriptor = %d',
        #   descriptor)

        # targets are mutable, so we can't use the cache
        pointer = vips_lib.vips_target_new_to_descriptor(descriptor)
        if pointer == ffi.NULL:
            raise Error(f"can't create output target from descriptor "
                        f'{descriptor}')

        return Target(pointer)

    @staticmethod
    def new_to_file(filename):
        """Make a new target to write to a file.

        Make a new target that will write to the named file. For example::

            target = pyvips.Target.new_to_file("myfile.jpg")

        You can pass this target to (for example) :meth:`write_to_target`.

        """

        # logger.debug('VipsTarget.new_to_file: filename = %s', filename)

        pointer = vips_lib.vips_target_new_to_file(_to_bytes(filename))
        if pointer == ffi.NULL:
            raise Error(f"can't create output target from filename {filename}")

        return Target(pointer)

    @staticmethod
    def new_to_memory():
        """Make a new target to write to an area of memory.

        Make a new target that will write to memory. For example::

            target = pyvips.Target.new_to_memory()

        You can pass this target to (for example) :meth:`write_to_target`.

        After writing to the target, fetch the bytes from the target object
        with `target.get("blob")`.

        """

        # logger.debug('VipsTarget.new_to_memory:')

        pointer = vips_lib.vips_target_new_to_memory()
        if pointer == ffi.NULL:
            raise Error("can't create output target from memory")

        return Target(pointer)


__all__ = ['Target']



# vim: set fileencoding=utf-8 :

import tempfile
import pytest

import pyvips
from helpers import JPEG_FILE, WEBP_FILE, temp_filename, skip_if_no


class TestConnections:
    @classmethod
    def setup_class(cls):
        cls.tempdir = tempfile.mkdtemp()

    @skip_if_no('jpegload')
    @pytest.mark.skipif(not pyvips.at_least_libvips(8, 9),
                        reason="requires libvips >= 8.9")
    def test_connection(self):
        source = pyvips.Source.new_from_file(JPEG_FILE)
        image = pyvips.Image.new_from_source(source, '', access='sequential')
        filename = temp_filename(self.tempdir, '.png')
        target = pyvips.Target.new_to_file(filename)
        image.write_to_target(target, '.png')

        image = pyvips.Image.new_from_file(JPEG_FILE, access='sequential')
        image2 = pyvips.Image.new_from_file(filename, access='sequential')

        assert (image - image2).abs().max() < 10

    @skip_if_no('jpegload')
    @pytest.mark.skipif(not pyvips.at_least_libvips(8, 9),
                        reason="requires libvips >= 8.9")
    def test_source_custom_no_seek(self):
        input_file = open(JPEG_FILE, "rb")

        def read_handler(size):
            return input_file.read(size)

        source = pyvips.SourceCustom()
        source.on_read(read_handler)

        image = pyvips.Image.new_from_source(source, '', access='sequential')
        image2 = pyvips.Image.new_from_file(JPEG_FILE, access='sequential')

        assert (image - image2).abs().max() == 0

    @skip_if_no('jpegload')
    @pytest.mark.skipif(not pyvips.at_least_libvips(8, 9),
                        reason="requires libvips >= 8.9")
    def test_source_custom(self):
        input_file = open(JPEG_FILE, "rb")

        def read_handler(size):
            return input_file.read(size)

        def seek_handler(offset, whence):
            input_file.seek(offset, whence)
            return input_file.tell()

        source = pyvips.SourceCustom()
        source.on_read(read_handler)
        source.on_seek(seek_handler)

        image = pyvips.Image.new_from_source(source, '',
                                             access='sequential')
        image2 = pyvips.Image.new_from_file(JPEG_FILE, access='sequential')

        assert (image - image2).abs().max() == 0

    @skip_if_no('jpegload')
    @pytest.mark.skipif(not pyvips.at_least_libvips(8, 9),
                        reason="requires libvips >= 8.9")
    def test_target_custom(self):
        filename = temp_filename(self.tempdir, '.png')
        output_file = open(filename, "w+b")

        def write_handler(chunk):
            return output_file.write(chunk)

        def end_handler():
            try:
                output_file.close()
            except IOError:
                return -1
            else:
                return 0

        target = pyvips.TargetCustom()
        target.on_write(write_handler)
        target.on_end(end_handler)

        image = pyvips.Image.new_from_file(JPEG_FILE, access='sequential')
        image.write_to_target(target, '.png')

        image = pyvips.Image.new_from_file(JPEG_FILE, access='sequential')
        image2 = pyvips.Image.new_from_file(filename, access='sequential')

        assert (image - image2).abs().max() == 0

    @skip_if_no('jpegload')
    @skip_if_no('tiffsave')
    @pytest.mark.skipif(not pyvips.at_least_libvips(8, 13),
                        reason="requires libvips >= 8.13")
    def test_target_custom_seek(self):
        filename = temp_filename(self.tempdir, '.png')
        output_file = open(filename, "w+b")

        def write_handler(chunk):
            return output_file.write(chunk)

        def read_handler(size):
            return output_file.read(size)

        def seek_handler(offset, whence):
            output_file.seek(offset, whence)
            return output_file.tell()

        def end_handler():
            try:
                output_file.close()
            except IOError:
                return -1
            else:
                return 0

        target = pyvips.TargetCustom()
        target.on_write(write_handler)
        target.on_read(read_handler)
        target.on_seek(seek_handler)
        target.on_end(end_handler)

        image = pyvips.Image.new_from_file(JPEG_FILE, access='sequential')
        image.write_to_target(target, '.tif')

        image = pyvips.Image.new_from_file(JPEG_FILE, access='sequential')
        image2 = pyvips.Image.new_from_file(filename, access='sequential')

        assert (image - image2).abs().max() == 0

    # test webp as well, since that maps the stream rather than using read

    @skip_if_no('webpload')
    @pytest.mark.skipif(not pyvips.at_least_libvips(8, 9),
                        reason="requires libvips >= 8.9")
    def test_source_custom_webp_no_seek(self):
        input_file = open(WEBP_FILE, "rb")

        def read_handler(size):
            return input_file.read(size)

        source = pyvips.SourceCustom()
        source.on_read(read_handler)

        image = pyvips.Image.new_from_source(source, '',
                                             access='sequential')
        image2 = pyvips.Image.new_from_file(WEBP_FILE, access='sequential')

        assert (image - image2).abs().max() == 0

    @skip_if_no('webpload')
    @pytest.mark.skipif(not pyvips.at_least_libvips(8, 9),
                        reason="requires libvips >= 8.9")
    def test_source_custom_webp(self):
        input_file = open(WEBP_FILE, "rb")

        def read_handler(size):
            return input_file.read(size)

        def seek_handler(offset, whence):
            input_file.seek(offset, whence)
            return input_file.tell()

        source = pyvips.SourceCustom()
        source.on_read(read_handler)
        source.on_seek(seek_handler)

        image = pyvips.Image.new_from_source(source, '',
                                             access='sequential')
        image2 = pyvips.Image.new_from_file(WEBP_FILE, access='sequential')

        assert (image - image2).abs().max() == 0



# errors from libvips

import logging

from pathlib import Path
from pyvips import ffi, vips_lib, glib_lib

logger = logging.getLogger(__name__)


def _to_bytes(x):
    """Convert to a byte string.

    Convert a Python unicode string or a pathlib.Path to a utf-8-encoded
    byte string. You must call this on strings you pass to libvips.

    """
    if isinstance(x, (str, Path)):
        # n.b. str also converts pathlib.Path objects
        x = str(x).encode('utf-8')

    return x


def _to_string(x):
    """Convert to a unicode string.

    If x is a byte string, assume it is utf-8 and decode to a Python unicode
    string. You must call this on text strings you get back from libvips.

    """
    if x == ffi.NULL:
        x = 'NULL'
    else:
        x = ffi.string(x)
        if isinstance(x, bytes):
            x = x.decode('utf-8')

    return x


def _to_string_copy(x):
    """Convert to a unicode string, and auto-free.

    As _to_string(), but also tag x as a pointer to a memory area that must
    be freed with g_free().

    """
    return _to_string(ffi.gc(x, glib_lib.g_free))


class Error(Exception):
    """An error from vips.

    Attributes:
        message (str): a high-level description of the error
        detail (str): a string with some detailed diagnostics

    """

    def __init__(self, message, detail=None):
        self.message = message
        if detail is None or detail == "":
            detail = _to_string(vips_lib.vips_error_buffer())
            vips_lib.vips_error_clear()
        self.detail = detail

        logger.debug('Error %s %s', self.message, self.detail)

    def __str__(self):
        return f'{self.message}\n  {self.detail}'


__all__ = [
    '_to_bytes',
    '_to_string',
    '_to_string_copy',
    'Error'
]



# wrap VipsObject

import logging

import pyvips
from pyvips import ffi, vips_lib, gobject_lib, Error, _to_bytes, _to_string

logger = logging.getLogger(__name__)


class VipsObject(pyvips.GObject):
    """Manage a VipsObject."""
    __slots__ = ('vobject', 'gobject')
    _pspec_cache = {}

    def __init__(self, pointer):
        # logger.debug('VipsObject.__init__: pointer = %s', pointer)
        super(VipsObject, self).__init__(pointer)
        self.vobject = ffi.cast('VipsObject*', pointer)
        self.gobject = ffi.cast('GObject*', pointer)

    @staticmethod
    def print_all():
        """Print all objects.

        Print a table of all active libvips objects. Handy for debugging.

        """
        vips_lib.vips_object_print_all()

    def _get_pspec(self, name):
        # logger.debug('VipsObject.get_typeof: self = %s, name = %s',
        #              str(self), name)

        # this is pretty slow, and used a lot, so we cache results
        # this cache makes the libvips test suite about 10% faster
        class_pointer = self.gobject.g_type_instance.g_class
        cache = VipsObject._pspec_cache
        if class_pointer not in cache:
            cache[class_pointer] = {}
        if name not in cache[class_pointer]:
            pspec = ffi.new('GParamSpec **')
            argument_class = ffi.new('VipsArgumentClass **')
            argument_instance = ffi.new('VipsArgumentInstance **')
            result = vips_lib.vips_object_get_argument(self.vobject,
                                                       _to_bytes(name),
                                                       pspec, argument_class,
                                                       argument_instance)

            if result != 0:
                return None

            cache[class_pointer][name] = pspec[0]

        return cache[class_pointer][name]

    def get_typeof(self, name):
        """Get the GType of a GObject property.

        This function returns 0 if the property does not exist.

        """

        # logger.debug('VipsObject.get_typeof: self = %s, name = %s',
        #              str(self), name)

        pspec = self._get_pspec(name)
        if pspec is None:
            # need to clear any error, this is horrible
            Error('')
            return 0

        return pspec.value_type

    def get_blurb(self, name):
        """Get the blurb for a GObject property."""

        c_str = gobject_lib.g_param_spec_get_blurb(self._get_pspec(name))
        return _to_string(c_str)

    def get(self, name):
        """Get a GObject property.

        The value of the property is converted to a Python value.

        """

        logger.debug('VipsObject.get: name = %s', name)

        pspec = self._get_pspec(name)
        if pspec is None:
            raise Error('Property not found.')
        gtype = pspec.value_type

        gv = pyvips.GValue()
        gv.set_type(gtype)
        go = ffi.cast('GObject *', self.pointer)
        gobject_lib.g_object_get_property(go, _to_bytes(name), gv.pointer)

        return gv.get()

    def set(self, name, value):
        """Set a GObject property.

        The value is converted to the property type, if possible.

        """

        logger.debug('VipsObject.set: name = %s, value = %s', name, value)

        gtype = self.get_typeof(name)

        gv = pyvips.GValue()
        gv.set_type(gtype)
        gv.set(value)
        go = ffi.cast('GObject *', self.pointer)
        gobject_lib.g_object_set_property(go, _to_bytes(name), gv.pointer)

    def set_string(self, string_options):
        """Set a series of properties using a string.

        For example::

            'fred=12, tile'
            '[fred=12]'

        """

        cstr = _to_bytes(string_options)
        result = vips_lib.vips_object_set_from_string(self.vobject, cstr)

        return result == 0

    def get_description(self):
        """Get the description of a GObject."""

        return _to_string(vips_lib.vips_object_get_description(self.vobject))


__all__ = ['VipsObject']



# vim: set fileencoding=utf-8 :

import pyvips
import pytest
from helpers import JPEG_FILE


class TestImage:
    def test_pagejoin(self):
        image = pyvips.Image.new_from_file(JPEG_FILE)
        many_page = image.pagejoin([image, image])
        assert many_page.get_page_height() == image.height
        assert many_page.height / many_page.get_page_height() == 3

    def test_pagesplit(self):
        image = pyvips.Image.new_from_file(JPEG_FILE)
        many_page = image.pagejoin([image, image])
        image_list = many_page.pagesplit()
        assert len(image_list) == 3

    def test_bandjoin(self):
        black = pyvips.Image.black(16, 16)
        a = black.draw_rect(1, 0, 0, 1, 1)
        b = black.draw_rect(2, 0, 0, 1, 1)
        c = black.draw_rect(3, 0, 0, 1, 1)
        im = a.bandjoin([b, c, a, b, c, a, b, c])

        assert im.width == 16
        assert im.height == 16
        assert im.bands == 9

        x = im.bandjoin([])
        assert x.bands == 9

    def test_bandslice(self):
        black = pyvips.Image.black(16, 16)
        a = black.draw_rect(1, 0, 0, 1, 1)
        b = black.draw_rect(2, 0, 0, 1, 1)
        c = black.draw_rect(3, 0, 0, 1, 1)
        d = black.draw_rect(4, 0, 0, 1, 1)
        e = black.draw_rect(5, 0, 0, 1, 1)
        f = black.draw_rect(6, 0, 0, 1, 1)

        im = black.bandjoin([a, b, c, d, e, f])

        seq = list(range(im.bands))

        x = im[1]

        assert x.bands == 1
        assert x(0, 0) == [seq[1]]

        x = im[1:4]

        assert x.bands == 3
        assert x(0, 0) == seq[1:4]

        x = im[1:4][::-1]

        assert x.bands == 3
        assert x(0, 0) == seq[1:4][::-1]

        x = im[-7]
        assert x(0, 0) == [seq[-7]]

        x = im[::-1]

        assert x(0, 0) == seq[::-1]

        x = im[4:6]

        assert x(0, 0) == seq[4:6]

        x = im[5:3:-1]

        assert x(0, 0) == seq[5:3:-1]

        x = im[2:4:2]

        assert x(0, 0) == seq[2:4:2]

        c = im[2:0:-2]

        assert x(0, 0) == seq[2:0:-2]

        x = im[::-2]

        assert x(0, 0) == seq[::-2]

        x = im[1::-2]

        assert x(0, 0) == seq[1::-2]

        x = im[:5:-2]

        assert x(0, 0) == seq[:5:-2]

        x = im[5:0:-2]

        assert x(0, 0) == seq[5:0:-2]

        x = im[-10:10]

        assert x(0, 0) == seq[-10:10]

        indices = [1, 2, 5]
        x = im[indices]

        assert x(0, 0) == [seq[i] for i in indices]

        indices = [1, 2, -7]
        x = im[indices]

        assert x(0, 0) == [seq[i] for i in indices]

        indices = [1]
        x = im[indices]

        assert x(0, 0) == [seq[1]]

        indices = [-1]
        x = im[indices]

        assert x(0, 0) == [seq[-1]]

        boolslice = [True, True, False, False, True, True, False]
        x = im[boolslice]
        assert x(0, 0) == [seq[i] for i, b in enumerate(boolslice) if b]

        with pytest.raises(IndexError):
            x = im[4:1]

        with pytest.raises(IndexError):
            x = im[-(im.bands + 1)]

        with pytest.raises(IndexError):
            x = im[im.bands]

        with pytest.raises(IndexError):
            empty = [False] * im.bands
            x = im[empty]

        with pytest.raises(IndexError):
            notenough = [True] * (im.bands - 1)
            x = im[notenough]

        with pytest.raises(IndexError):
            toomany = [True] * (im.bands + 1)
            x = im[toomany]

        with pytest.raises(IndexError):
            empty = []
            x = im[empty]

        with pytest.raises(IndexError):
            oob = [2, 3, -8]
            x = im[oob]

        with pytest.raises(IndexError):
            mixed = [True, 1, True, 2, True, 3, True]
            x = im[mixed]

        with pytest.raises(IndexError):
            wrongtypelist = ['a', 'b', 'c']
            x = im[wrongtypelist]

        with pytest.raises(IndexError):
            wrongargtype = dict(a=1, b=2)
            x = im[wrongargtype]

    def test_invalidate(self):
        try:
            import numpy as np
        except ImportError:
            pytest.skip('numpy not available')

        a = np.zeros((1, 1))
        p = pyvips.Image.new_from_memory(a.data, 1, 1, 1, 'double')
        v = p(0, 0)
        assert v == [0]
        a[0, 0] = 1
        v = p(0, 0)
        assert v == [0]
        p.invalidate()
        v = p(0, 0)
        assert v == [1]

    def test_to_numpy(self):
        try:
            import numpy as np
        except ImportError:
            pytest.skip('numpy not available')

        xy = pyvips.Image.xyz(4, 5)

        # using __array__ interface:

        from pyvips.vimage import TYPESTR_TO_FORMAT
        for typestr, format in TYPESTR_TO_FORMAT.items():
            this_xy = xy.cast(format)
            if typestr == '|b1':
                yx = np.asarray(this_xy, dtype=typestr)
            else:
                yx = np.array(this_xy)

            assert yx.dtype == np.dtype(typestr)
            assert yx.shape == (5, 4, 2)
            assert all(yx.max(axis=(0, 1)) == np.array([3, 4], dtype=typestr))

        x, y = xy
        x_iy = pyvips.Image.complexform(x, y)
        x_iy_dp = x_iy.cast('dpcomplex')

        a = np.array(x_iy)
        assert a.shape == (5, 4)
        assert a.dtype == np.dtype('complex64')
        assert a[4, 3].item() == 3+4j

        a = np.array(x_iy_dp)
        assert a.shape == (5, 4)
        assert a.dtype == np.dtype('complex128')
        assert a[4, 3].item() == 3+4j

        xyxyxy = xy.bandjoin([xy, xy])
        a = np.array(xyxyxy)
        assert a.shape == (5, 4, 6)

        # axes collapsing:
        xyxyxy_col = xyxyxy.crop(3, 0, 1, 5)
        assert xyxyxy_col.width == 1
        assert xyxyxy_col.height == 5
        a = np.array(xyxyxy_col)
        assert a.shape == (5, 1, 6)

        xyxyxy_row = xyxyxy.crop(0, 4, 4, 1)
        assert xyxyxy_row.width == 4
        assert xyxyxy_row.height == 1
        a = np.array(xyxyxy_row)
        assert a.shape == (1, 4, 6)

        xyxyxy_px = xyxyxy.crop(3, 4, 1, 1)
        assert xyxyxy_px.width == 1
        assert xyxyxy_px.height == 1
        a = np.array(xyxyxy_px)
        assert a.shape == (1, 1, 6)

        a = np.array(xyxyxy_px[0])
        assert a.ndim == 0

        # automatic conversion to array on ufunc application:

        d = np.diff(xy.cast('short'), axis=1)
        assert (d[..., 0] == 1).all()
        assert (d[..., 1] == 0).all()

        # tests of .numpy() method:

        a = pyvips.Image.xyz(256, 1)[0].numpy().squeeze()

        assert all(a == np.arange(256))

    def test_scipy(self):
        try:
            import numpy as np
        except ImportError:
            pytest.skip('numpy not available')

        try:
            from scipy import ndimage
        except ImportError:
            pytest.skip('scipy not available')

        black = pyvips.Image.black(16, 16)
        a = black.draw_rect(1, 0, 0, 1, 1)

        d = ndimage.distance_transform_edt(a == 0)
        assert np.allclose(d[-1, -1], (2 * 15 ** 2) ** 0.5)

    def test_torch(self):
        try:
            import numpy as np
        except ImportError:
            pytest.skip('numpy not available')

        try:
            import torch
        except ImportError:
            pytest.skip('torch not available')

        # torch to Image:
        x = torch.outer(torch.arange(10), torch.arange(5))
        with pytest.raises(ValueError):
            # no vips format for int64
            im = pyvips.Image.new_from_array(x)

        x = x.float()

        im = pyvips.Image.new_from_array(x)
        assert im.width == 5
        assert im.height == 10
        assert im(4, 9) == [36]
        assert im.format == 'float'

        # Image to torch:
        im = pyvips.Image.zone(5, 5)
        t = torch.from_numpy(np.asarray(im))
        assert t[2, 2] == 1.

    def test_from_numpy(self):
        try:
            import numpy as np
        except ImportError:
            pytest.skip('numpy not available')

        a = np.indices((5, 4)).transpose(1, 2, 0)  # (5, 4, 2)

        with pytest.raises(ValueError):
            # no way in for int64
            yx = pyvips.Image.new_from_array(a)

        from pyvips.vimage import TYPESTR_TO_FORMAT

        for typestr, format in TYPESTR_TO_FORMAT.items():
            if typestr == '|b1':
                continue
            a = np.indices((5, 4), dtype=typestr).transpose(1, 2, 0)
            yx = pyvips.Image.new_from_array(a)
            assert yx.format == format

        a = np.zeros((2, 2, 2, 2))

        with pytest.raises(ValueError):
            # too many dimensions
            im = pyvips.Image.new_from_array(a)

        a = np.ones((2, 2, 2), dtype=bool)
        im = pyvips.Image.new_from_array(a)
        assert im.max() == 255

        a = np.ones((2, 2, 2), dtype='S8')
        with pytest.raises(ValueError):
            # no way in for strings
            im = pyvips.Image.new_from_array(a)

        # test handling of strided data
        a = np.ones((1000, 1000, 3), dtype='uint8')[::10, ::10]
        im = pyvips.Image.new_from_array(a)
        assert im.width == 100
        assert im.height == 100
        assert im.bands == 3

        class FakeArray(object):
            @property
            def __array_interface__(self):
                return {'shape': (1, 1, 1),
                        'typestr': '|u1',
                        'version': 3}

        with pytest.raises(TypeError):
            # Handle evil objects that don't behave like ndarrays
            im = pyvips.Image.new_from_array(FakeArray())

    def test_tolist(self):
        im = pyvips.Image.complexform(*pyvips.Image.xyz(3, 4))

        assert im.tolist()[-1][-1] == 2+3j

        im = im.cast('dpcomplex')

        assert im.tolist()[-1][-1] == 2+3j

        lst = [[1, 2, 3], [4, 5, 6]]

        im = pyvips.Image.new_from_array(lst)

        assert im.tolist() == lst

        assert im.cast('float').tolist() == lst
        assert im.cast('complex').tolist() == lst

    def test_from_PIL(self):
        try:
            import PIL.Image
        except ImportError:
            pytest.skip('PIL not available')

        pim = PIL.Image.new('RGB', (42, 23))

        im = pyvips.Image.new_from_array(pim)
        assert im.format == 'uchar'
        assert im.interpretation == 'srgb'
        assert im.width == pim.width
        assert im.height == pim.height
        assert im.min() == 0
        assert im.max() == 0



import logging

import pyvips
from pyvips import ffi, vips_lib, Error, _to_bytes, _to_string, GValue, \
    type_map, type_from_name, nickname_find, at_least_libvips

logger = logging.getLogger(__name__)

# values for VipsArgumentFlags
_REQUIRED = 1
_CONSTRUCT = 2
_SET_ONCE = 4
_SET_ALWAYS = 8
_INPUT = 16
_OUTPUT = 32
_DEPRECATED = 64
_MODIFY = 128

# for VipsOperationFlags
_OPERATION_DEPRECATED = 8


class Introspect(object):
    """Build introspection data for operations.

    Make an operation, introspect it, and build a structure representing
    everything we know about it.

    """
    __slots__ = ('description', 'flags', 'details',
                 'required_input', 'optional_input',
                 'required_output', 'optional_output',
                 'doc_optional_input', 'doc_optional_output',
                 'member_x', 'method_args')

    def __init__(self, operation_name):
        op = Operation.new_from_name(operation_name)

        self.description = op.get_description()
        self.flags = vips_lib.vips_operation_get_flags(op.pointer)

        # build a list of constructor arg [name, flags] pairs in arg order
        arguments = []

        def add_args(name, flags):
            if (flags & _CONSTRUCT) != 0:
                # libvips uses '-' to separate parts of arg names, but we
                # need '_' for Python
                name = name.replace('-', '_')
                arguments.append([name, flags])

        if at_least_libvips(8, 7):
            p_names = ffi.new('char**[1]')
            p_flags = ffi.new('int*[1]')
            p_n_args = ffi.new('int[1]')
            result = vips_lib.vips_object_get_args(op.vobject,
                                                   p_names, p_flags, p_n_args)
            if result != 0:
                raise Error('unable to get arguments from operation')
            p_names = p_names[0]
            p_flags = p_flags[0]
            n_args = p_n_args[0]

            for i in range(0, n_args):
                add_args(_to_string(p_names[i]), p_flags[i])
        else:
            def add_construct(self, pspec, argument_class,
                              argument_instance, a, b):
                add_args(_to_string(pspec.name), argument_class.flags)
                return ffi.NULL

            cb = ffi.callback('VipsArgumentMapFn', add_construct)
            vips_lib.vips_argument_map(op.vobject, cb, ffi.NULL, ffi.NULL)

        # logger.debug('arguments = %s', self.arguments)

        # build a hash from arg name to detailed arg information
        self.details = {}
        for name, flags in arguments:
            self.details[name] = {
                "name": name,
                "flags": flags,
                "blurb": op.get_blurb(name),
                "type": op.get_typeof(name)
            }

        # lists of arg names by category
        self.required_input = []
        self.optional_input = []
        self.required_output = []
        self.optional_output = []

        # same, but with deprecated args filtered out ... this is the set we
        # show in documentation
        self.doc_optional_input = []
        self.doc_optional_output = []

        for name, flags in arguments:
            if ((flags & _INPUT) != 0 and
                    (flags & _REQUIRED) != 0 and
                    (flags & _DEPRECATED) == 0):
                self.required_input.append(name)

                # required inputs which we MODIFY are also required outputs
                if (flags & _MODIFY) != 0:
                    self.required_output.append(name)

            if ((flags & _OUTPUT) != 0 and
                    (flags & _REQUIRED) != 0 and
                    (flags & _DEPRECATED) == 0):
                self.required_output.append(name)

            # deprecated optional args get on to the main arg lists, but are
            # filtered from the documented set
            if ((flags & _INPUT) != 0 and
                    (flags & _REQUIRED) == 0):
                self.optional_input.append(name)

                if (flags & _DEPRECATED) == 0:
                    self.doc_optional_input.append(name)

            if ((flags & _OUTPUT) != 0 and
                    (flags & _REQUIRED) == 0):
                self.optional_output.append(name)

                if (flags & _DEPRECATED) == 0:
                    self.doc_optional_output.append(name)

        # find the first required input image arg, if any ... that will be self
        self.member_x = None
        for name in self.required_input:
            details = self.details[name]
            if details['type'] == GValue.image_type:
                self.member_x = name
                break

        # method args are required args, but without the image they are a
        # method on
        if self.member_x is not None:
            self.method_args = list(self.required_input)
            self.method_args.remove(self.member_x)
        else:
            self.method_args = self.required_input

    # a hash mapping operation names to introspection data
    _introspect_cache = {}

    @classmethod
    def get(cls, operation_name):
        if operation_name not in cls._introspect_cache:
            cls._introspect_cache[operation_name] = Introspect(operation_name)

        return cls._introspect_cache[operation_name]


# search an array with a predicate, recursing into subarrays as we see them
# used to find the match_image for an operation
def _find_inside(pred, thing):
    if pred(thing):
        return thing

    if isinstance(thing, list) or isinstance(thing, tuple):
        for x in thing:
            result = _find_inside(pred, x)

            if result is not None:
                return result

    return None


class Operation(pyvips.VipsObject):
    """Call libvips operations.

    This class wraps the libvips VipsOperation class.

    """
    __slots__ = ()

    # cache nickname -> docstring here
    _docstring_cache = {}

    def __init__(self, pointer):
        # logger.debug('Operation.__init__: pointer = %s', pointer)
        super(Operation, self).__init__(pointer)

    @staticmethod
    def new_from_name(operation_name):
        vop = vips_lib.vips_operation_new(_to_bytes(operation_name))
        if vop == ffi.NULL:
            raise Error(f'no such operation {operation_name}')
        return Operation(vop)

    def set(self, name, flags, match_image, value):
        # if the object wants an image and we have a constant, _imageize it
        #
        # if the object wants an image array, _imageize any constants in the
        # array
        if match_image:
            gtype = self.get_typeof(name)

            if gtype == pyvips.GValue.image_type:
                value = pyvips.Image._imageize(match_image, value)
            elif gtype == pyvips.GValue.array_image_type:
                value = [pyvips.Image._imageize(match_image, x)
                         for x in value]

        # MODIFY args need to be copied before they are set
        if (flags & _MODIFY) != 0:
            # logger.debug('copying MODIFY arg %s', name)
            # make sure we have a unique copy
            value = value.copy().copy_memory()

        super(Operation, self).set(name, value)

    @staticmethod
    def call(operation_name, *args, **kwargs):
        """Call a libvips operation.

        Use this method to call any libvips operation. For example::

            black_image = pyvips.Operation.call('black', 10, 10)

        See the Introduction for notes on how this works.

        """

        logger.debug('VipsOperation.call: operation_name = %s', operation_name)
        # logger.debug('VipsOperation.call: args = %s, kwargs =%s',
        #              args, kwargs)

        intro = Introspect.get(operation_name)

        if len(intro.required_input) != len(args):
            raise Error(f'{operation_name} needs {len(intro.required_input)} '
                        f'arguments, but {len(args)} given')

        op = Operation.new_from_name(operation_name)

        # set any string options before any args so they can't be
        # overridden
        string_options = kwargs.pop('string_options', '')
        if not op.set_string(string_options):
            raise Error(f'unable to call {operation_name}')

        # the first image argument is the thing we expand constants to
        # match ... look inside tables for images, since we may be passing
        # an array of images as a single param
        match_image = _find_inside(lambda x: isinstance(x, pyvips.Image),
                                   args)

        logger.debug('VipsOperation.call: match_image = %s', match_image)

        # collect a list of all input references here
        # we can't use a set because set elements are unique under "==", and
        # Python checks memoryview equality with hash functions, not pointer
        # equality
        references = []

        # does a list contain an element using pointer equality
        # we can't use "in" since that uses "==", which means hash equality
        def contains(array, x):
            for y in array:
                if id(x) == id(y):
                    return True
            return False

        def add_reference(x):
            if isinstance(x, pyvips.Image):
                for i in x._references:
                    if not contains(references, i):
                        references.append(i)
            return False

        # set required input args
        for name, value in zip(intro.required_input, args):
            _find_inside(add_reference, value)
            op.set(name, intro.details[name]['flags'], match_image, value)

        # set any optional args
        for name in kwargs:
            if (name not in intro.optional_input and
                    name not in intro.optional_output):
                raise Error(f'{operation_name} does not support optional '
                            f'argument {name}')

            value = kwargs[name]
            details = intro.details[name]

            if (details['flags'] & _DEPRECATED) != 0:
                logger.info('%s argument %s is deprecated',
                            operation_name, name)

            _find_inside(add_reference, value)
            op.set(name, details['flags'], match_image, value)

        # build operation
        vop = vips_lib.vips_cache_operation_build(op.pointer)
        if vop == ffi.NULL:
            vips_lib.vips_object_unref_outputs(op.vobject)
            raise Error(f'unable to call {operation_name}')
        op = Operation(vop)

        # attach all input refs to output x
        def set_reference(x):
            if isinstance(x, pyvips.Image):
                x._references.extend(references)
            return False

        # fetch required output args (plus modified input images)
        result = []
        for name in intro.required_output:
            value = op.get(name)
            _find_inside(set_reference, value)
            result.append(value)

        # fetch optional output args
        opts = {}
        for name in intro.optional_output:
            if name in kwargs:
                value = op.get(name)
                _find_inside(set_reference, value)
                opts[name] = value

        if len(opts) > 0:
            result.append(opts)

        vips_lib.vips_object_unref_outputs(op.vobject)

        if len(result) == 0:
            result = None
        elif len(result) == 1:
            result = result[0]

        logger.debug('VipsOperation.call: result = %s', result)

        return result

    @staticmethod
    def generate_docstring(operation_name):
        """Make a google-style docstring.

        This is used to generate help() output.

        """

        # we cache these to save regeneration
        if operation_name in Operation._docstring_cache:
            return Operation._docstring_cache[operation_name]

        intro = Introspect.get(operation_name)
        if (intro.flags & _OPERATION_DEPRECATED) != 0:
            raise Error('No such operator.',
                        f'operator "{operation_name}" is deprecated')

        result = intro.description[0].upper() + intro.description[1:] + '.\n\n'
        result += 'Example:\n'

        result += '   ' + ', '.join(intro.required_output) + ' = '
        if intro.member_x is not None:
            result += intro.member_x + '.' + operation_name + '('
        else:
            result += 'pyvips.Image.' + operation_name + '('

        args = []
        args += intro.method_args
        args += [x + '=' + GValue.gtype_to_python(intro.details[x]['type'])
                 for x in intro.doc_optional_input]
        args += [x + '=bool'
                 for x in intro.doc_optional_output]
        result += ", ".join(args) + ')\n'

        def argstr(name):
            details = intro.details[name]
            return (u'    {0} ({1}): {2}\n'.
                    format(name,
                           GValue.gtype_to_python(details['type']),
                           details['blurb']))

        result += '\nReturns:\n'
        for name in intro.required_output:
            result += argstr(name)

        result += '\nArgs:\n'
        if intro.member_x is not None:
            result += argstr(intro.member_x)
        for name in intro.method_args:
            result += argstr(name)

        if len(intro.doc_optional_input) > 0:
            result += '\nKeyword args:\n'
            for name in intro.doc_optional_input:
                result += argstr(name)

        if len(intro.doc_optional_output) > 0:
            result += '\nOther Parameters:\n'
            for name in intro.doc_optional_output:
                result += argstr(name)

        result += '\nRaises:\n    :class:`.Error`\n'

        # add to cache to save building again
        Operation._docstring_cache[operation_name] = result

        return result

    @staticmethod
    def generate_sphinx(operation_name):
        """Make a sphinx-style docstring.

        This is used to generate the off-line docs.

        """

        intro = Introspect.get(operation_name)
        if (intro.flags & _OPERATION_DEPRECATED) != 0:
            raise Error('No such operator.',
                        f'operator "{operation_name}" is deprecated')

        if intro.member_x is not None:
            result = '.. method:: '
        else:
            result = '.. staticmethod:: '
        args = []
        args += intro.method_args
        args += [x + '=' + GValue.gtype_to_python(intro.details[x]['type'])
                 for x in intro.doc_optional_input]
        args += [x + '=bool'
                 for x in intro.doc_optional_output]
        result += operation_name + '(' + ", ".join(args) + ')\n\n'

        result += intro.description[0].upper() + \
            intro.description[1:] + '.\n\n'

        result += 'Example:\n'
        result += '    '
        if len(intro.required_output) > 0:
            result += ', '.join(intro.required_output) + ' = '
        if intro.member_x is not None:
            result += intro.member_x + "." + operation_name + '('
        else:
            result += 'pyvips.Image.' + operation_name + '('
        args = []
        args += intro.method_args
        args += [x + '=' + GValue.gtype_to_python(intro.details[x]['type'])
                 for x in intro.doc_optional_input]
        result += ', '.join(args)
        result += ')\n\n'

        for name in intro.method_args + intro.doc_optional_input:
            details = intro.details[name]
            result += f':param {name}: {details["blurb"]}\n'
            result += (f':type {name}: '
                       f'{GValue.gtype_to_python(details["type"])}\n')
        for name in intro.doc_optional_output:
            result += (f':param {name}: '
                       f'enable output: {intro.details[name]["blurb"]}\n')
            result += f':type {name}: bool\n'

        output_types = [GValue.gtype_to_python(intro.details[name]['type'])
                        for name in intro.required_output]
        if len(output_types) == 1:
            output_type = output_types[0]
        else:
            output_type = 'list[' + ', '.join(output_types) + ']'

        if len(intro.doc_optional_output) > 0:
            output_types += ['Dict[str, mixed]']
            output_type += ' or list[' + ', '.join(output_types) + ']'

        result += ':rtype: ' + output_type + '\n'
        result += ':raises Error:\n'

        return result

    @staticmethod
    def generate_sphinx_all():
        """Generate sphinx documentation.

        This generates a .rst file for all auto-generated image methods. Use it
        to regenerate the docs with something like::

            $ python -c \
"import pyvips; pyvips.Operation.generate_sphinx_all()" > x

        And copy-paste the file contents into doc/vimage.rst in the appropriate
        place.

        """

        # these names are aliased
        alias = ["crop"]
        alias_gtypes = {}
        for name in alias:
            gtype = pyvips.type_find("VipsOperation", name)
            alias_gtypes[gtype] = name

        # all names we can generate docstrings for
        all_names = []

        def add_name(gtype, a, b):
            if gtype in alias_gtypes:
                name = alias_gtypes[gtype]
            else:
                name = nickname_find(gtype)

            try:
                Operation.generate_sphinx(name)
                all_names.append(name)
            except Error:
                pass

            type_map(gtype, add_name)

            return ffi.NULL

        type_map(type_from_name('VipsOperation'), add_name)

        all_names.sort()

        # remove operations we have to wrap by hand
        exclude = ['scale', 'ifthenelse', 'bandjoin', 'bandrank']
        all_names = [x for x in all_names if x not in exclude]

        # Output summary table
        print('.. class:: Image')
        print('   :noindex:\n')
        print('   .. rubric:: Methods\n')
        print('   .. autosummary::')
        print('      :nosignatures:\n')
        for name in all_names:
            print(f'      ~{name}')
        print()

        # Output docs
        print()
        for name in all_names:
            docstr = Operation.generate_sphinx(name)
            docstr = docstr.replace('\n', '\n      ')
            print('   ' + docstr)


def cache_set_max(mx):
    """Set the maximum number of operations libvips will cache."""
    vips_lib.vips_cache_set_max(mx)


def cache_set_max_mem(mx):
    """Limit the operation cache by memory use."""
    vips_lib.vips_cache_set_max_mem(mx)


def cache_set_max_files(mx):
    """Limit the operation cache by number of open files."""
    vips_lib.vips_cache_set_max_files(mx)


def cache_set_trace(trace):
    """Turn on libvips cache tracing."""
    vips_lib.vips_cache_set_trace(trace)


def cache_get_max():
    """Get the maximum number of operations libvips will cache."""
    return vips_lib.vips_cache_get_max()


def cache_get_size():
    """Get the current size of the operations cache."""
    return vips_lib.vips_cache_get_size()


def cache_get_max_mem():
    """Get the operation cache limit by memory use."""
    return vips_lib.vips_cache_get_max_mem()


def cache_get_max_files():
    """Get the operation cache limit by number of open files."""
    return vips_lib.vips_cache_get_max_files()


def block_untrusted_set(state):
    """Set the block state for all untrusted operations."""
    if at_least_libvips(8, 13):
        vips_lib.vips_block_untrusted_set(state)


def operation_block_set(name, state):
    """Set the block state for a named operation."""
    if at_least_libvips(8, 13):
        vips_lib.vips_operation_block_set(_to_bytes(name), state)


__all__ = [
    'Introspect',
    'Operation',
    'cache_set_max',
    'cache_set_max_mem',
    'cache_set_max_files',
    'cache_set_trace',
    'cache_get_max',
    'cache_get_max_mem',
    'cache_get_max_files',
    'cache_get_size',
    'block_untrusted_set',
    'operation_block_set'
]



# wrap VipsImage

import numbers
import struct

import pyvips
from pyvips import ffi, glib_lib, vips_lib, Error, _to_bytes, \
    _to_string, _to_string_copy, GValue, at_least_libvips, Introspect


# either a single number, or a table of numbers
def _is_pixel(value):
    return (isinstance(value, numbers.Number) or
            (isinstance(value, list) and not
            isinstance(value, pyvips.Image)))


# test for rectangular array of something
def _is_2D(array):
    if not isinstance(array, list):
        return False

    for x in array:
        if not isinstance(x, list):
            return False
        if len(x) != len(array[0]):
            return False

    return True


# apply a function to a thing, or map over a list
# we often need to do something like (1.0 / other) and need to work for lists
# as well as scalars
def _smap(func, x):
    if isinstance(x, list):
        return list(map(func, x))
    else:
        return func(x)


def _call_enum(image, other, base, operation):
    if _is_pixel(other):
        return pyvips.Operation.call(base + '_const', image, operation, other)
    else:
        return pyvips.Operation.call(base, image, other, operation)


def _run_cmplx(fn, image):
    """Run a complex function on a non-complex image.

    The image needs to be complex, or have an even number of bands. The input
    can be int, the output is always float or double.
    """
    original_format = image.format

    if image.format != 'complex' and image.format != 'dpcomplex':
        if image.bands % 2 != 0:
            raise Error('not an even number of bands')

        if image.format != 'float' and image.format != 'double':
            image = image.cast('float')

        if image.format == 'double':
            new_format = 'dpcomplex'
        else:
            new_format = 'complex'

        image = image.copy(format=new_format, bands=image.bands / 2)

    image = fn(image)

    if original_format != 'complex' and original_format != 'dpcomplex':
        if image.format == 'dpcomplex':
            new_format = 'double'
        else:
            new_format = 'float'

        image = image.copy(format=new_format, bands=image.bands * 2)

    return image


# decorator to set docstring
def _add_doc(name):
    try:
        docstring = pyvips.Operation.generate_docstring(name)
    except Error:
        docstring = None

    def _doc(func):
        func.__doc__ = docstring
        return func

    return _doc


# decorator to set deprecated
def _deprecated(note):
    def _dep(func):
        func.__deprecated__ = note
        return func

    return _dep


# Rules for the mapping between numpy typestrings and libvips formats
# b1: bool. Above u1 so that rev. map is uchar->u1
TYPESTR_TO_FORMAT = {'|b1': 'uchar',
                     '|u1': 'uchar',
                     '|i1': 'char',
                     '<u2': 'ushort',
                     '<i2': 'short',
                     '<u4': 'uint',
                     '<i4': 'int',
                     '<f4': 'float',
                     '<c8': 'complex',
                     '<f8': 'double',
                     '<c16': 'dpcomplex'}

# Rules for the mapping between libvips formats and numpy typestrings
FORMAT_TO_TYPESTR = dict((v, k) for k, v in TYPESTR_TO_FORMAT.items())

# see https://docs.python.org/3/library/struct.html
FORMAT_TO_PYFORMAT = {'uchar': 'B',
                      'char': 'c',
                      'ushort': 'H',
                      'short': 'h',
                      'uint': 'I',
                      'int': 'i',
                      'float': 'f',
                      'double': 'd',
                      'complex': 'f',   # n.b. no support for complex in C
                      'dpcomplex': 'd'}


def _guess_interpretation(bands, format):
    """Return a best-guess interpretation string based on bands and libvips
    format.

    Args:
        bands (int): Number of bands in the image.
        format (str): libvips format string.

    Returns:
        str: the best-guess libvips interpretation string.

    This function is a helper for :class:`.Image` creation from arrays when the
    user specifies 'auto' for `interpretation`.

    The heuristic used here can be defined in the following table::

        | bands | format    | interpretation | reason       |
        | ----- | --------- | -------------- | ------------ |
        | 1     | uchar     | b-w            | L            |
        | 2     | uchar     | b-w            | LA           |
        | 3     | uchar     | rgb            | RGB          |
        | 4     | uchar     | rgb            | RGBA         |
        | 5+    | uchar     | multiband      | spectral     |
        | 1     | char      | matrix         | conv. kernel |
        | 2+    | char      | multiband      |              |
        | 1     | ushort    | grey16         | L            |
        | 2     | ushort    | grey16         | LA           |
        | 3     | ushort    | rgb16          | RGB          |
        | 4     | ushort    | rgb16          | RGBA         |
        | 5+    | ushort    | multiband      | spectral     |
        | any   | short     | multiband      |              |
        | 1     | float     | b-w            | L            |
        | 2     | float     | b-w            | LA           |
        | 3     | float     | scrgb          | RGB          |
        | 4     | float     | scrgb          | RGBA         |
        | 5+    | float     | multiband      | spectral     |
        | 1     | double    | b-w            | L            |
        | 2     | double    | b-w            | LA           |
        | 3     | double    | scrgb          | RGB          |
        | 4     | double    | scrgb          | RGBA         |
        | 5+    | double    | multiband      | spectral     |
        | any   | complex   | fourier        | FFT          |
        | any   | dpcomplex | fourier        | FFT          |

    """

    if format not in FORMAT_TO_TYPESTR:
        raise ValueError(f'Unknown format: {format}')
    if not isinstance(bands, int) or bands < 1:
        raise ValueError('Number of bands must be a positive integer.')

    # the default
    interp = 'multiband'

    if format == 'uchar':
        if bands in [1, 2]:
            interp = 'b-w'
        elif bands in [3, 4]:
            interp = 'rgb'
    elif format == 'char':
        if bands == 1:
            interp = 'matrix'
    elif format == 'ushort':
        if bands in [1, 2]:
            interp = 'grey16'
        elif bands in [3, 4]:
            interp = 'rgb16'
    elif format == 'short':
        if bands == 1:
            interp = 'matrix'
    elif format in ['float', 'double']:
        if bands in [1, 2]:
            interp = 'b-w'
        elif bands in [3, 4]:
            interp = 'scrgb'
    elif format in ['complex', 'dpcomplex']:
        interp = 'fourier'

    return interp


# metaclass for Image ... getattr on this implements the class methods
class ImageType(type):
    def __getattr__(cls, name):
        # logger.debug('ImageType.__getattr__ %s', name)

        # does the method exist in libvips?
        try:
            # this will throw an exception if not
            Introspect.get(name)
        except Error:
            # we need to throw this exception for missing methods, eg. numpy
            # checks for this
            raise AttributeError

        @_add_doc(name)
        def call_function(*args, **kwargs):
            return pyvips.Operation.call(name, *args, **kwargs)

        return call_function


class Image(pyvips.VipsObject, metaclass=ImageType):
    """Wrap a VipsImage object.

    """
    __slots__ = ('_references',)

    # private static

    @staticmethod
    def _imageize(self, value):
        # careful! self can be None if value is a 2D array
        if isinstance(value, Image):
            return value
        elif _is_2D(value):
            return Image.new_from_array(value)
        else:
            return self.new_from_image(value)

    def __init__(self, pointer):
        # a list of other objects which this object depends on and which need
        # to be kept alive
        # we can't use a set because set elements are unique under "==", and
        # Python checks memoryview equality with hash functions, not pointer
        # equality
        self._references = []
        # logger.debug('Image.__init__: pointer = %s', pointer)
        super(Image, self).__init__(pointer)

    # constructors

    @staticmethod
    def new_from_file(vips_filename, **kwargs):
        """Load an image from a file.

        This method can load images in any format supported by vips. The
        filename can include load options, for example::

            image = pyvips.Image.new_from_file('fred.jpg[shrink=2]')

        You can also supply options as keyword arguments, for example::

            image = pyvips.Image.new_from_file('fred.jpg', shrink=2)

        The full set of options available depend upon the load operation that
        will be executed. Try something like::

            $ vips jpegload

        at the command-line to see a summary of the available options for the
        JPEG loader.

        Loading is fast: only enough of the image is loaded to be able to fill
        out the header. Pixels will only be decompressed when they are needed.

        Args:
            vips_filename (str): The disc file to load the image from, with
                optional appended arguments.

        All loaders support at least the following options:

        Keyword args:
            memory (bool): If set True, load the image via memory rather than
                via a temporary disc file. See :meth:`.new_temp_file` for
                notes on where temporary files are created. Small images are
                loaded via memory by default, use ``VIPS_DISC_THRESHOLD`` to
                set the definition of small.
            access (Access): Hint the expected access pattern for the image.
            fail (bool): If set True, the loader will fail with an error on
                the first serious error in the file. By default, libvips
                will attempt to read everything it can from a damaged image.

        Returns:
            A new :class:`.Image`.

        Raises:
            :class:`.Error`

        """
        vips_filename = _to_bytes(vips_filename)
        pointer = vips_lib.vips_filename_get_filename(vips_filename)
        filename = _to_string_copy(pointer)

        pointer = vips_lib.vips_filename_get_options(vips_filename)
        options = _to_string_copy(pointer)

        pointer = vips_lib.vips_foreign_find_load(vips_filename)
        if pointer == ffi.NULL:
            raise Error(f'unable to load from file {vips_filename}')
        name = _to_string(pointer)

        return pyvips.Operation.call(name, filename,
                                     string_options=options, **kwargs)

    @staticmethod
    def new_from_buffer(data, options, **kwargs):
        """Load a formatted image from memory.

        This behaves exactly as :meth:`new_from_file`, but the image is
        loaded from the memory object rather than from a file. The memory
        object can be anything that supports the Python buffer protocol.

        Args:
            data (array, bytearray, bytes, buffer): The memory object to
                load the image from.
            options (str): Load options as a string. Use ``""`` for no options.

        All loaders support at least the following options:

        Keyword args:
            access (Access): Hint the expected access pattern for the image.
            fail (bool): If set True, the loader will fail with an error on the
                first serious error in the image. By default, libvips will
                attempt to read everything it can from a damaged image.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        pointer = vips_lib.vips_foreign_find_load_buffer(data, len(data))
        if pointer == ffi.NULL:
            raise Error('unable to load from buffer')
        name = _to_string(pointer)

        return pyvips.Operation.call(name, data,
                                     string_options=options, **kwargs)

    @staticmethod
    def new_from_list(array, scale=1.0, offset=0.0):
        """Create an image from a list or list of lists.

        A new one-band image with :class:`BandFormat` ``'double'`` pixels is
        created from the array. These image are useful with the libvips
        convolution operator :meth:`Image.conv`.

        Args:
            array (list[list[float]]): Create the image from these values.
                1D arrays become a single row of pixels.
            scale (float): Default to 1.0. What to divide each pixel by after
                convolution.  Useful for integer convolution masks.
            offset (float): Default to 0.0. What to subtract from each pixel
                after convolution.  Useful for integer convolution masks.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        if not _is_2D(array):
            array = [array]

        height = len(array)
        width = len(array[0])

        n = width * height
        a = ffi.new('double[]', n)
        for y in range(0, height):
            for x in range(0, width):
                a[x + y * width] = array[y][x]

        vi = vips_lib.vips_image_new_matrix_from_array(width, height, a, n)
        if vi == ffi.NULL:
            raise Error('unable to make image from matrix')
        image = pyvips.Image(vi)

        image.set_type(GValue.gdouble_type, 'scale', scale)
        image.set_type(GValue.gdouble_type, 'offset', offset)

        return image

    @classmethod
    def new_from_array(cls, obj, scale=1.0, offset=0.0, interpretation=None):
        """Create a new Image from a list or an array-like object.

        Array-like objects are those which define `__array_interface__` or
        `__array__`. For details about the array interface, see `The Array
        Interface
        <https://numpy.org/doc/stable/reference/arrays.interface.html>`_.

        If `__array_interface__` is not available, `__array__` is used as a
        fallback.

        The behavior for input objects with different dimensions is summarized
        as::

            | array ndim | array shape | Image w | Image h | Image bands |
            |------------|-------------|---------|---------|-------------|
            | 0          | ()          | 1       | 1       | 1           |
            | 1          | (W,)        | W       | 1       | 1           |
            | 2          | (H, W)      | W       | H       | 1           |
            | 3          | (H, W, C)   | W       | H       | C           |


        Args:
            obj (list or object width `__array_interface__` or `__array__`):
                The object to convert to an image.

                If the input object is a list, `Image.new_from_list` is used
                with the given `scale` and `offset`

                If the input object is an array-like object, a new image is
                created from the object's data and shape.  The memory is shared
                except in the following cases:

                - The object's memory is not contiguous.  In this case, a copy
                  is made by attempting to call the object's `tobytes()` method
                  or its `tostring()` method.

                - The object is an array of bools, in which case it is
                  converted to a pyvips uchar image with True values becoming
                  255 and False values becoming 0.

            scale (float): Default to 1.0. **Ignored for non-list inputs**.
                What to divide each pixel by after convolution.  Useful for
                integer convolution masks.
            offset (float): Default to 0.0.  **Ignored for non-list inputs**.
                What to subtract from each pixel after convolution.  Useful for
                integer convolution masks.
            interpretation (str, optional): **Ignored for list inputs** The
                libvips interpretation of the array. If None, the
                interpretation defaults to the pyvips one for
                `Image.new_from_memory`.

                If 'auto', a heuristic is used to determine a best-guess
                interpretation as defined in the `_guess_interpretation`
                function.

                Must be one of None, 'auto', 'error', 'multiband', 'b-w',
                'histogram', 'xyz', 'lab', 'cmyk', 'labq', 'rgb', 'cmc', 'lch',
                'labs', 'srgb', 'yxy', 'fourier', 'rgb16', 'grey16', 'matrix',
                'scrgb', or 'hsv'

        Returns:
            The new image.

        See Also:
            :func:`_guess_interpretation`
        """
        if isinstance(obj, list):
            return cls.new_from_list(obj, scale, offset)

        if hasattr(obj, '__array_interface__'):
            a = obj.__array_interface__
            shape = a['shape']
            typestr = a['typestr']
            ndim = len(shape)

            # strides is optional
            strides = a.get('strides', None)

            if ndim > 3:
                raise ValueError('array has more than 3 dimensions')
            if typestr not in TYPESTR_TO_FORMAT:
                raise ValueError(f'conversion from {typestr} not supported')

            if ndim == 0:
                width = 1
                height = 1
                bands = 1
            elif ndim == 1:
                width = shape[0]
                height = 1
                bands = 1
            elif ndim == 2:
                height, width = shape
                bands = 1
            elif ndim == 3:
                height, width, bands = shape

            format = TYPESTR_TO_FORMAT[typestr]

            if strides is None and hasattr(obj, 'data'):
                data = obj.data
            else:
                # To obtain something with a contiguous memory layout
                if hasattr(obj, 'tobytes'):
                    data = obj.tobytes()
                elif hasattr(obj, 'tostring'):
                    data = obj.tostring()
                else:
                    raise TypeError('object has no .tobytes or .tostring')

            im = cls.new_from_memory(
                data,
                width,
                height,
                bands,
                format
            )

            if typestr == '|b1':
                # special case for bool: true in vips is uchar(255)
                im = im.ifthenelse(255, 0)

            if interpretation is not None:
                if interpretation == 'auto':
                    interpretation = _guess_interpretation(bands, format)

                im = im.copy(interpretation=interpretation)

            return im

        elif hasattr(obj, '__array__'):
            # make it into something that *does* define __array_interface__
            return cls.new_from_array(obj.__array__())
        else:
            raise TypeError('does not define __array_interface__ '
                            'or __array__')

    @staticmethod
    def new_from_memory(data, width, height, bands, format):
        """Wrap an image around a memory array.

        Wraps an Image around an area of memory containing a C-style array. For
        example, if the ``data`` memory array contains four bytes with the
        values 1, 2, 3, 4, you can make a one-band, 2x2 uchar image from
        it like this::

            image = Image.new_from_memory(data, 2, 2, 1, 'uchar')

        A reference is kept to the data object, so it will not be
        garbage-collected until the returned image is garbage-collected.

        This method is useful for efficiently transferring images from PIL or
        NumPy into libvips.

        See :meth:`.write_to_memory` for the opposite operation.

        Use :meth:`.copy` to set other image attributes.

        Args:
            data (bytes): A memoryview or buffer object.
            width (int): Image width in pixels.
            height (int): Image height in pixels.
            bands (int): Number of bands.
            format (BandFormat): Band format.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        format_value = GValue.to_enum(GValue.format_type, format)
        pointer = ffi.from_buffer(data)
        nbytes = data.nbytes if hasattr(data, 'nbytes') else len(data)
        vi = vips_lib.vips_image_new_from_memory(pointer,
                                                 nbytes,
                                                 width, height, bands,
                                                 format_value)
        if vi == ffi.NULL:
            raise Error('unable to make image from memory')

        image = pyvips.Image(vi)

        # keep a secret ref to the underlying object .. this reference will be
        # inherited by things that in turn depend on us, so the memory we are
        # using will not be freed
        image._references.append(data)

        return image

    @staticmethod
    def new_from_source(source, options, **kwargs):
        """Load a formatted image from a source.

        This behaves exactly as :meth:`new_from_file`, but the image is
        loaded from the source rather than from a file.

        Args:
            source (Source): The source to load the image from.
            options (str): Load options as a string. Use ``""`` for no options.

        All loaders support at least the following options:

        Keyword args:
            access (Access): Hint the expected access pattern for the image.
            fail (bool): If set True, the loader will fail with an error on the
                first serious error in the image. By default, libvips will
                attempt to read everything it can from a damaged image.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        pointer = vips_lib.vips_foreign_find_load_source(source.pointer)
        if pointer == ffi.NULL:
            raise Error('unable to load from source')
        name = _to_string(pointer)

        image = pyvips.Operation.call(name, source,
                                      string_options=options, **kwargs)

        # keep a secret ref to the source object .. we need that to stay
        # alive, since it might be a custom source that triggers a python
        # callback
        image._references.append(source)

        return image

    @staticmethod
    def new_temp_file(format):
        """Make a new temporary image.

        Returns an image backed by a temporary file. When written to with
        :func:`Image.write`, a temporary file will be created on disc in the
        specified format. When the image is closed, the file will be deleted
        automatically.

        The file is created in the temporary directory. This is set with
        the environment variable ``TMPDIR``. If this is not set, then on
        Unix systems, vips will default to ``/tmp``. On Windows, vips uses
        ``GetTempPath()`` to find the temporary directory.

        vips uses ``g_mkstemp()`` to make the temporary filename. They
        generally look something like ``"vips-12-EJKJFGH.v"``.

        Args:
            format (str): The format for the temp file, for example
                ``"%s.v"`` for a vips format file. The ``%s`` is
                substituted by the file path.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        vi = vips_lib.vips_image_new_temp_file(_to_bytes(format))
        if vi == ffi.NULL:
            raise Error('unable to make temp file')

        return pyvips.Image(vi)

    def new_from_image(self, value):
        """Make a new image from an existing one.

        A new image is created which has the same size, format, interpretation
        and resolution as ``self``, but with every pixel set to ``value``.

        Args:
            value (float, list[float]): The value for the pixels. Use a
                single number to make a one-band image; use an array constant
                to make a many-band image.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        pixel = (Image.black(1, 1) + value).cast(self.format)
        image = pixel.embed(0, 0, self.width, self.height,
                            extend='copy')
        image = image.copy(interpretation=self.interpretation,
                           xres=self.xres,
                           yres=self.yres,
                           xoffset=self.xoffset,
                           yoffset=self.yoffset)

        return image

    def copy_memory(self):
        """Copy an image to memory.

        A large area of memory is allocated, the image is rendered to that
        memory area, and a new image is returned which wraps that large memory
        area.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        vi = vips_lib.vips_image_copy_memory(self.pointer)
        if vi == ffi.NULL:
            raise Error('unable to copy to memory')

        return pyvips.Image(vi)

    # writers

    def write_to_file(self, vips_filename, **kwargs):
        """Write an image to a file on disc.

        This method can save images in any format supported by vips. The format
        is selected from the filename suffix. The filename can include embedded
        save options, see :func:`Image.new_from_file`.

        For example::

            image.write_to_file('fred.jpg[Q=95]')

        You can also supply options as keyword arguments, for example::

            image.write_to_file('fred.jpg', Q=95)

        The full set of options available depend upon the load operation that
        will be executed. Try something like::

            $ vips jpegsave

        at the command-line to see a summary of the available options for the
        JPEG saver.

        Args:
            vips_filename (str): The disc file to save the image to, with
                optional appended arguments.

        Other arguments depend upon the save operation.

        Returns:
            None

        Raises:
            :class:`.Error`

        """
        vips_filename = _to_bytes(vips_filename)
        pointer = vips_lib.vips_filename_get_filename(vips_filename)
        filename = _to_string_copy(pointer)

        pointer = vips_lib.vips_filename_get_options(vips_filename)
        options = _to_string_copy(pointer)

        pointer = vips_lib.vips_foreign_find_save(vips_filename)
        if pointer == ffi.NULL:
            raise Error(f'unable to write to file {vips_filename}')
        name = _to_string(pointer)

        return pyvips.Operation.call(name, self, filename,
                                     string_options=options, **kwargs)

    def write_to_buffer(self, format_string, **kwargs):
        """Write an image to memory.

        This method can save images in any format supported by vips. The format
        is selected from the suffix in the format string. This can include
        embedded save options, see :func:`Image.write_to_file`.

        For example::

            data = image.write_to_buffer('.jpg[Q=95]')

        You can also supply options as keyword arguments, for example::

            data = image.write_to_buffer('.jpg', Q=95)

        The full set of options available depend upon the load operation that
        will be executed. Try something like::

            $ vips jpegsave_buffer

        at the command-line to see a summary of the available options for the
        JPEG saver.

        Args:
            format_string (str): The suffix, plus any string-form arguments.

        Other arguments depend upon the save operation.

        Returns:
            A byte string.

        Raises:
            :class:`.Error`

        """
        format_string = _to_bytes(format_string)

        filename = vips_lib.vips_filename_get_filename(format_string)
        filename = ffi.gc(filename, glib_lib.g_free)

        pointer = vips_lib.vips_filename_get_options(format_string)
        options = _to_string_copy(pointer)

        pointer = ffi.NULL
        if at_least_libvips(8, 9):
            vips_lib.vips_error_freeze()
            pointer = vips_lib.vips_foreign_find_save_target(filename)
            vips_lib.vips_error_thaw()

        if pointer != ffi.NULL:
            name = _to_string(pointer)
            target = pyvips.Target.new_to_memory()
            pyvips.Operation.call(name, self, target,
                                  string_options=options, **kwargs)
            buffer = target.get("blob")
        else:
            pointer = vips_lib.vips_foreign_find_save_buffer(filename)
            if pointer == ffi.NULL:
                raise Error('unable to write to buffer')

            name = _to_string(pointer)
            buffer = pyvips.Operation.call(name, self,
                                           string_options=options, **kwargs)

        return buffer

    def write_to_target(self, target, format_string, **kwargs):
        """Write an image to a target.

        This method will write the image to the target in the format
        specified in the suffix in the format string. This can include
        embedded save options, see :func:`Image.write_to_file`.

        For example::

            image.write_to_target(target, '.jpg[Q=95]')

        You can also supply options as keyword arguments, for example::

            image.write_to_target(target, '.jpg', Q=95)

        The full set of options available depend upon the save operation that
        will be executed. Try something like::

            $ vips jpegsave_target

        at the command-line to see a summary of the available options for the
        JPEG saver.

        Args:
            target (Target): The target to write the image to
            format_string (str): The suffix, plus any string-form arguments.

        Other arguments depend upon the save operation.

        Returns:
            None

        Raises:
            :class:`.Error`

        """
        format_string = _to_bytes(format_string)

        pointer = vips_lib.vips_filename_get_options(format_string)
        options = _to_string_copy(pointer)

        pointer = vips_lib.vips_foreign_find_save_target(format_string)
        if pointer == ffi.NULL:
            raise Error('unable to write to target')
        name = _to_string(pointer)

        return pyvips.Operation.call(name, self, target,
                                     string_options=options, **kwargs)

    def write_to_memory(self):
        """Write the image to a large memory array.

        A large area of memory is allocated, the image is rendered to that
        memory array, and the array is returned as a buffer.

        For example, if you have a 2x2 uchar image containing the bytes 1, 2,
        3, 4, read left-to-right, top-to-bottom, then::

            buf = image.write_to_memory()

        will return a four byte buffer containing the values 1, 2, 3, 4.

        Returns:
            buffer

        Raises:
            :class:`.Error`

        """
        psize = ffi.new('size_t *')
        pointer = vips_lib.vips_image_write_to_memory(self.pointer, psize)
        if pointer == ffi.NULL:
            raise Error('unable to write to memory')
        pointer = ffi.gc(pointer, glib_lib.g_free)

        return ffi.buffer(pointer, psize[0])

    def write(self, other):
        """Write an image to another image.

        This function writes ``self`` to another image. Use something like
        :func:`Image.new_temp_file` to make an image that can be written to.

        Args:
            other (Image): The :class:`Image` to write to,

        Returns:
            None

        Raises:
            :class:`.Error`

        """
        result = vips_lib.vips_image_write(self.pointer, other.pointer)
        if result != 0:
            raise Error('unable to write to image')

    def invalidate(self):
        """Drop caches on an image, and any downstream images.

        This method drops all pixel caches on an image and on all downstream
        images. Any operations which depend on this image, directly or
        indirectly, are also dropped from the libvips operation cache.

        This method can be useful if you wrap a libvips image around an area
        of memory with :meth:`.new_from_memory` and then change some bytes
        without libvips knowing.

        Returns:
            None

        """
        vips_lib.vips_image_invalidate_all(self.pointer)

    def set_progress(self, progress):
        """Enable progress reporting on an image.

        When progress reporting is enabled, evaluation of the most downstream
        image from this image will report progress using the ::preeval, ::eval,
        and ::posteval signals.

        """
        vips_lib.vips_image_set_progress(self.pointer, progress)

    def set_kill(self, kill):
        """Kill processing of an image.

        Use this to kill evaluation of an image. You can call it from one of
        the progress signals, for example. See :meth:`.set_progress`.

        """
        vips_lib.vips_image_set_kill(self.pointer, kill)

    # get/set metadata

    def get_typeof(self, name):
        """Get the GType of an item of metadata.

        Fetch the GType of a piece of metadata, or 0 if the named item does not
        exist. See :class:`GValue`.

        Args:
            name (str): The name of the piece of metadata to get the type of.

        Returns:
            The ``GType``, or 0.

        Raises:
            None

        """
        # on libvips before 8.5, property types must be fetched separately,
        # since built-in enums were reported as ints
        if not at_least_libvips(8, 5):
            gtype = super(Image, self).get_typeof(name)
            if gtype != 0:
                return gtype

        return vips_lib.vips_image_get_typeof(self.pointer, _to_bytes(name))

    def get(self, name):
        """Get an item of metadata.

        Fetches an item of metadata as a Python value. For example::

            orientation = image.get('orientation')

        would fetch the image orientation.

        Args:
            name (str): The name of the piece of metadata to get.

        Returns:
            The metadata item as a Python value.

        Raises:
            :class:`.Error`

        """
        # with old libvips, we must fetch properties (as opposed to
        # metadata) via VipsObject
        if not at_least_libvips(8, 5):
            gtype = super(Image, self).get_typeof(name)
            if gtype != 0:
                return super(Image, self).get(name)

        gv = GValue()
        result = vips_lib.vips_image_get(self.pointer, _to_bytes(name),
                                         gv.pointer)
        if result != 0:
            raise Error(f'unable to get {name}')

        return gv.get()

    def get_fields(self):
        """Get a list of all the metadata fields on an image.

        Returns:
            [string]

        """
        names = []

        if at_least_libvips(8, 5):
            array = vips_lib.vips_image_get_fields(self.pointer)
            i = 0
            while array[i] != ffi.NULL:
                name = _to_string(array[i])
                names.append(name)
                glib_lib.g_free(array[i])
                i += 1
            glib_lib.g_free(array)

        return names

    def set_type(self, gtype, name, value):
        """Set the type and value of an item of metadata.

        Sets the type and value of an item of metadata. Any old item of the
        same name is removed. See :class:`GValue` for types.

        Args:
            gtype (int): The GType of the metadata item to create.
            name (str): The name of the piece of metadata to create.
            value (mixed): The value to set as a Python value. It is
                converted to the ``gtype``, if possible.

        Returns:
            None

        Raises:
            None

        """
        gv = GValue()
        gv.set_type(gtype)
        gv.set(value)
        vips_lib.vips_image_set(self.pointer, _to_bytes(name), gv.pointer)

    def set(self, name, value):
        """Set the value of an item of metadata.

        Sets the value of an item of metadata. The metadata item must already
        exist.

        Args:
            name (str): The name of the piece of metadata to set the value of.
            value (mixed): The value to set as a Python value. It is
                converted to the type of the metadata item, if possible.

        Returns:
            None

        Raises:
            :class:`.Error`

        """
        gtype = self.get_typeof(name)
        if gtype == 0:
            raise Error(f'metadata item {name} does not exist -'
                        ' use set_type() to create and set')
        self.set_type(gtype, name, value)

    def remove(self, name):
        """Remove an item of metadata.

        The named metadata item is removed.

        Args:
            name (str): The name of the piece of metadata to remove.

        Returns:
            None

        Raises:
            None

        """
        return vips_lib.vips_image_remove(self.pointer, _to_bytes(name)) != 0

    def tolist(self):
        """Return a single-band image as a list of lists.

        Returns:
            list of lists of values

        """
        if not self.bands == 1:
            raise NotImplementedError('tolist only implemented for ' +
                                      'single-band images')

        is_complex = self.format in ['complex', 'dpcomplex']

        row_els = self.width if not is_complex else 2 * self.width

        rowfmt = f'{row_els}{FORMAT_TO_PYFORMAT[self.format]}'
        buf = self.write_to_memory()

        lst = [list(row) for row in struct.iter_unpack(rowfmt, buf)]

        if is_complex:
            lst = [[r[i] + 1j*r[i+1] for i in range(0, row_els, 2)]
                   for r in lst]

        return lst

    # Following will be relevant if we expose the buffer interface:
    # @property
    # def __array_interface__(self):
    #     """Return a numpy-standard __array_interface__ dictionary.
    #
    #     This is used by some libraries (PIL, e.g.) to interpret the image
    #     as an array.
    #
    #     See https://numpy.org/doc/stable/reference/arrays.interface.html for
    #     more info.
    #
    #     """
    #     if self.bands == 1:
    #         shape = (self.height, self.width)
    #     else:
    #         shape = (self.height, self.width, self.bands)
    #
    #     typestr = FORMAT_TO_TYPESTR[self.format]
    #
    #     interface = {
    #         'shape': shape,
    #         'strides': None,
    #         'descr': [('', typestr)],
    #         'typestr': typestr,
    #         'version': 3
    #     }
    #
    #     return interface

    def __array__(self, dtype=None):
        """Conversion to a NumPy array.

        Args:
            dtype (str or numpy dtype, optional) The dtype to use for the
                numpy array. If None, the default dtype of the image is used
                as defined the global `FORMAT_TO_TYPESTR` dictionary.

        Returns:
            numpy.ndarray: The array representation of the image.
                * Single-band images lose their channel axis.
                * Single-pixel single-band images are converted to a 0D array.

        See https://numpy.org/devdocs/user/basics.dispatch.html for more
        details.

        This enables a :class:`Image` to be used where a numpy array is
        expected, including in plotting and conversion to other array
        container types (pytorch, JAX, dask, etc.), for example.

        `numpy` is a runtime dependency of this function.

        See Also `Image.new_from_array` for the inverse operation. #TODO
        """
        import numpy as np

        arr = (
            np.frombuffer(self.write_to_memory(),
                          dtype=FORMAT_TO_TYPESTR[self.format])
            .reshape(self.height, self.width, self.bands)
        )

        if self.bands == 1:
            # flatten single-band images
            arr = arr.squeeze(axis=-1)

            if self.width == 1 and self.height == 1:
                # flatten single-pixel single-band images
                arr = arr.squeeze(axis=(0, 1))

        if dtype is not None:
            arr = arr.astype(dtype)

        return arr

    def numpy(self, dtype=None):
        """Convenience function to allow numpy conversion to be at the end
        of a method chain.

        This mimics the behavior of pytorch: ``arr = im.op1().op2().numpy()``

        numpy is a runtime dependency of this function.

        Args:
            dtype (str or numpy dtype): The dtype to use for the numpy array.
                If None, the default dtype of the image is used.

        Returns:
            numpy.ndarray: The image as a numpy array.

        See Also:

            - :meth:`.__array__`
            - `FORMAT_TO_TYPESTR`: Global dictionary mapping libvips format
              strings to numpy dtype strings.
        """
        return self.__array__(dtype=dtype)

    def __repr__(self):
        if (
            self.interpretation == "matrix"
            and self.width < 20
            and self.height < 20
            and self.bands == 1
        ):
            array = self.tolist()
            return repr(array)
        else:
            return (f'<pyvips.Image {self.width}x{self.height} {self.format}, '
                    f'{self.bands} bands, {self.interpretation}>')

    def __getattr__(self, name):
        """Divert unknown names to libvips.

        Unknown attributes are first looked up in the image properties as
        accessors, for example::

            width = image.width

        and then in the libvips operation table, where they become method
        calls, for example::

            new_image = image.invert()

        Use :func:`get` to fetch image metadata.

        A ``__getattr__`` on the metatype lets you call static members in the
        same way.

        Args:
            name (str): The name of the piece of metadata to get.

        Returns:
            Mixed.

        Raises:
            :class:`.Error`

        """
        # logger.debug('Image.__getattr__ %s', name)

        # scale and offset have default values
        if name == 'scale':
            if self.get_typeof('scale') != 0:
                return self.get('scale')
            else:
                return 1.0

        if name == 'offset':
            if self.get_typeof('offset') != 0:
                return self.get('offset')
            else:
                return 0.0

        # look up in props first (but not metadata)
        if super(Image, self).get_typeof(name) != 0:
            return super(Image, self).get(name)

        # does the method exist in libvips?
        try:
            # this will throw an exception if not
            Introspect.get(name)
        except Error:
            # we need to throw this exception for missing methods, eg. numpy
            # checks for this
            raise AttributeError

        @_add_doc(name)
        def call_function(*args, **kwargs):
            return pyvips.Operation.call(name, self, *args, **kwargs)

        return call_function

    # compatibility methods

    @_deprecated('use Image.get() instead')
    def get_value(self, name):
        return self.get(name)

    @_deprecated('use Image.set() instead')
    def set_value(self, name, value):
        self.set(name, value)

    @_deprecated('use Image.scale instead')
    def get_scale(self):
        return self.scale

    @_deprecated('use Image.offset instead')
    def get_offset(self):
        return self.offset

    # support with in the most trivial way

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def __getitem__(self, arg):
        """Overload [] to pull out band elements from an image.

        The following arguments types are accepted:

        * int::

            green = rgb_image[1]

          Will make a new one-band image from band 1 (the middle band).

        * slice::

            last_two = rgb_image[1:]
            last_band = rgb_image[-1]
            middle_few = multiband[1:-2]
            reversed  = multiband[::-1]
            every_other = multiband[::2]
            other_every_other = multiband[1::2]

        * list of int::

            # list of integers
            desired_bands = [1, 2, 2, -1]
            four_band = multiband[desired_bands]

        * list of bool::

            wanted_bands = [True, False, True, True, False]
            three_band = five_band[wanted_bands]

        In the case of integer or slice arguments, the semantics of slicing
        exactly match those of slicing `range(self.bands)`.

        In the case of list arguments, the semantics match those of numpy's
        extended slicing syntax. Thus, lists of booleans must have as many
        elements as there are bands in the image.

        """
        if isinstance(arg, int):
            # raises IndexError for bad integer indices
            i = range(self.bands)[arg]
            n = 1
        elif isinstance(arg, slice):
            band_seq = range(self.bands)[arg]
            if len(band_seq) == 0:
                raise IndexError('empty slice')

            i = band_seq[0]
            n = len(band_seq)
        elif isinstance(arg, list):
            if not arg:
                raise IndexError('empty list')

            # n.b. We don't use isinstance because isinstance(True, int)
            # is True!
            if not (all(type(x) == int for x in arg) or  # noqa: E721
                    all(type(x) == bool for x in arg)):  # noqa: E721
                raise IndexError('list must contain only ints or only bools')

            if isinstance(arg[0], bool):
                if len(arg) != self.bands:
                    raise IndexError('boolean index must have same ' +
                                     'length as bands')

                band_seq = [i for i, x in enumerate(arg) if x]
                n = len(band_seq)
                if n == 0:
                    raise IndexError('empty boolean index')
                if n == 1:
                    i = band_seq[0]
            else:
                all_bands = range(self.bands)
                # raises IndexError for bad int indices
                band_seq = [all_bands[i] for i in arg]

                n = len(band_seq)
                if n == 1:
                    i = band_seq[0]
        else:
            raise IndexError('argument must be an int, slice, ' +
                             'list of ints, or list of bools')

        if n == 1:
            # int or 1-element selection
            return self.extract_band(i)
        elif isinstance(arg, slice) and arg.step == 1:
            # sequential multi-band slice
            return self.extract_band(i, n=n)
        else:
            # nonsequential slice, list of ints, or list of bools
            bands = [self.extract_band(x) for x in band_seq]
            return bands[0].bandjoin(bands[1:])

    # overload () to mean fetch pixel
    def __call__(self, x, y):
        """Fetch a pixel value.

        Args:
            x (int): The x coordinate to fetch.
            y (int): The y coordinate to fetch.

        Returns:
            Pixel as an array of floating point numbers.

        Raises:
            :class:`.Error`

        """
        return self.getpoint(x, y)

    # operator overloads

    def __add__(self, other):
        if isinstance(other, pyvips.Image):
            return self.add(other)
        else:
            return self.linear(1, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, pyvips.Image):
            return self.subtract(other)
        else:
            return self.linear(1, _smap(lambda x: -1 * x, other))

    def __rsub__(self, other):
        return self.linear(-1, other)

    def __mul__(self, other):
        if isinstance(other, Image):
            return self.multiply(other)
        else:
            return self.linear(other, 0)

    def __rmul__(self, other):
        return self.__mul__(other)

    # a / const has always been a float in vips, so div and truediv are the
    # same
    def __div__(self, other):
        if isinstance(other, pyvips.Image):
            return self.divide(other)
        else:
            return self.linear(_smap(lambda x: 1.0 / x, other), 0)

    def __rdiv__(self, other):
        return (self ** -1) * other

    def __truediv__(self, other):
        return self.__div__(other)

    def __rtruediv__(self, other):
        return self.__rdiv__(other)

    def __floordiv__(self, other):
        if isinstance(other, pyvips.Image):
            return self.divide(other).floor()
        else:
            return self.linear(_smap(lambda x: 1.0 / x, other), 0).floor()

    def __rfloordiv__(self, other):
        return ((self ** -1) * other).floor()

    def __mod__(self, other):
        if isinstance(other, pyvips.Image):
            return self.remainder(other)
        else:
            return self.remainder_const(other)

    def __pow__(self, other):
        return _call_enum(self, other, 'math2', 'pow')

    def __rpow__(self, other):
        return _call_enum(self, other, 'math2', 'wop')

    def __abs__(self):
        return self.abs()

    def __lshift__(self, other):
        return _call_enum(self, other, 'boolean', 'lshift')

    def __rshift__(self, other):
        return _call_enum(self, other, 'boolean', 'rshift')

    def __and__(self, other):
        return _call_enum(self, other, 'boolean', 'and')

    def __rand__(self, other):
        return self.__and__(other)

    def __or__(self, other):
        return _call_enum(self, other, 'boolean', 'or')

    def __ror__(self, other):
        return self.__or__(other)

    def __xor__(self, other):
        return _call_enum(self, other, 'boolean', 'eor')

    def __rxor__(self, other):
        return self.__xor__(other)

    def __neg__(self):
        return -1 * self

    def __pos__(self):
        return self

    def __invert__(self):
        return self ^ -1

    def __gt__(self, other):
        return _call_enum(self, other, 'relational', 'more')

    def __ge__(self, other):
        return _call_enum(self, other, 'relational', 'moreeq')

    def __lt__(self, other):
        return _call_enum(self, other, 'relational', 'less')

    def __le__(self, other):
        return _call_enum(self, other, 'relational', 'lesseq')

    def __eq__(self, other):
        # _eq version allows comparison to None
        if other is None:
            return False

        return _call_enum(self, other, 'relational', 'equal')

    def __ne__(self, other):
        # _eq version allows comparison to None
        if other is None:
            return True

        return _call_enum(self, other, 'relational', 'noteq')

    def floor(self):
        """Return the largest integral value not greater than the argument."""
        return self.round('floor')

    def ceil(self):
        """Return the smallest integral value not less than the argument."""
        return self.round('ceil')

    def rint(self):
        """Return the nearest integral value."""
        return self.round('rint')

    def bandand(self):
        """AND image bands together."""
        return self.bandbool('and')

    def bandor(self):
        """OR image bands together."""
        return self.bandbool('or')

    def bandeor(self):
        """EOR image bands together."""
        return self.bandbool('eor')

    def bandsplit(self):
        """Split an n-band image into n separate images."""
        return [x for x in self]

    def bandjoin(self, other):
        """Append a set of images or constants bandwise."""
        if not isinstance(other, list):
            other = [other]

        # guard against empty list
        if not other:
            return self

        # if [other] is all numbers, we can use bandjoin_const
        non_number = next((x for x in other
                           if not isinstance(x, numbers.Number)),
                          None)

        if non_number is None:
            return self.bandjoin_const(other)
        else:
            return pyvips.Operation.call('bandjoin', [self] + other)

    def atan2(self, other):
        return _call_enum(self, other, 'math2', 'atan2')

    def get_n_pages(self):
        """Get the number of pages in an image file, or 1.

        This is the number of pages in the file the image was loaded from, not
        the number of pages in the image.

        To get the number of pages in an image, divide the image height by
        the page height.
        """
        if at_least_libvips(8, 8):
            return vips_lib.vips_image_get_n_pages(self.pointer)
        else:
            return self.get("n-pages") \
                if self.get_typeof("n-pages") != 0 else 1

    def get_page_height(self):
        """Get the page height in a many-page image, or height.
        """
        if at_least_libvips(8, 8):
            return vips_lib.vips_image_get_page_height(self.pointer)
        else:
            if self.get_typeof("page-height") != 0:
                page_height = self.get("page-height")
            else:
                page_height = self.height

            if page_height > 0 and \
                page_height <= self.height and \
                    self.height % page_height == 0:
                return page_height
            else:
                return self.height

    def pagesplit(self):
        """Split an N-page image into a list of N separate images.
        """
        page_height = self.get_page_height()
        return [self.crop(0, y, self.width, page_height)
                for y in range(0, self.height, page_height)]

    def pagejoin(self, other):
        """Join a set of pages vertically to make a multipage image.

        Also sets the page-height property on the result.
        """
        if not isinstance(other, list):
            other = [other]

        n_pages = 1 + len(other)

        image = Image.arrayjoin([self] + other, across=1)
        image = image.copy()
        image.set_type(GValue.gint_type, 'page-height', image.height / n_pages)

        return image

    def composite(self, other, mode, **kwargs):
        """Composite a set of images with a set of modes."""
        if not isinstance(other, list):
            other = [other]
        if not isinstance(mode, list):
            mode = [mode]

        # modes are VipsBlendMode enums, but we have to pass as array of int --
        # we need to map str->int by hand
        mode = [GValue.to_enum(GValue.blend_mode_type, x) for x in mode]

        return pyvips.Operation.call('composite', [self] + other, mode,
                                     **kwargs)

    def bandrank(self, other, **kwargs):
        """Band-wise rank filter a set of images."""
        if not isinstance(other, list):
            other = [other]

        return pyvips.Operation.call('bandrank', [self] + other, **kwargs)

    def maxpos(self):
        """Return the coordinates of the image maximum."""
        v, opts = self.max(x=True, y=True)
        x = opts['x']
        y = opts['y']
        return v, x, y

    def minpos(self):
        """Return the coordinates of the image minimum."""
        v, opts = self.min(x=True, y=True)
        x = opts['x']
        y = opts['y']
        return v, x, y

    def real(self):
        """Return the real part of a complex image."""
        return self.complexget('real')

    def imag(self):
        """Return the imaginary part of a complex image."""
        return self.complexget('imag')

    def polar(self):
        """Return an image converted to polar coordinates."""
        return _run_cmplx(lambda x: x.complex('polar'), self)

    def rect(self):
        """Return an image converted to rectangular coordinates."""
        return _run_cmplx(lambda x: x.complex('rect'), self)

    def conj(self):
        """Return the complex conjugate of an image."""
        return self.complex('conj')

    def sin(self):
        """Return the sine of an image in degrees."""
        return self.math('sin')

    def cos(self):
        """Return the cosine of an image in degrees."""
        return self.math('cos')

    def tan(self):
        """Return the tangent of an image in degrees."""
        return self.math('tan')

    def asin(self):
        """Return the inverse sine of an image in degrees."""
        return self.math('asin')

    def acos(self):
        """Return the inverse cosine of an image in degrees."""
        return self.math('acos')

    def atan(self):
        """Return the inverse tangent of an image in degrees."""
        return self.math('atan')

    def sinh(self):
        """Return the hyperbolic sine of an image in radians."""
        return self.math('sinh')

    def cosh(self):
        """Return the hyperbolic cosine of an image in radians."""
        return self.math('cosh')

    def tanh(self):
        """Return the hyperbolic tangent of an image in radians."""
        return self.math('tanh')

    def asinh(self):
        """Return the inverse hyperbolic sine of an image in radians."""
        return self.math('asinh')

    def acosh(self):
        """Return the inverse hyperbolic cosine of an image in radians."""
        return self.math('acosh')

    def atanh(self):
        """Return the inverse hyperbolic tangent of an image in radians."""
        return self.math('atanh')

    def log(self):
        """Return the natural log of an image."""
        return self.math('log')

    def log10(self):
        """Return the log base 10 of an image."""
        return self.math('log10')

    def exp(self):
        """Return e ** pixel."""
        return self.math('exp')

    def exp10(self):
        """Return 10 ** pixel."""
        return self.math('exp10')

    def erode(self, mask):
        """Erode with a structuring element."""
        return self.morph(mask, 'erode')

    def dilate(self, mask):
        """Dilate with a structuring element."""
        return self.morph(mask, 'dilate')

    def median(self, size):
        """size x size median filter."""
        return self.rank(size, size, (size * size) / 2)

    def fliphor(self):
        """Flip horizontally."""
        return self.flip('horizontal')

    def flipver(self):
        """Flip vertically."""
        return self.flip('vertical')

    def rot90(self):
        """Rotate 90 degrees clockwise."""
        return self.rot('d90')

    def rot180(self):
        """Rotate 180 degrees."""
        return self.rot('d180')

    def rot270(self):
        """Rotate 270 degrees clockwise."""
        return self.rot('d270')

    def hasalpha(self):
        """True if the image has an alpha channel."""
        return vips_lib.vips_image_hasalpha(self.pointer)

    def addalpha(self):
        """Add an alpha channel."""
        if self.interpretation == 'grey16' or self.interpretation == 'rgb16':
            max_alpha = 65535
        elif self.interpretation == 'scrgb':
            max_alpha = 1.0
        else:
            max_alpha = 255
        return self.bandjoin(max_alpha)

    # we need different _imageize rules for this operator ... we need to
    # _imageize in1 and in2 to match each other first
    @_add_doc('ifthenelse')
    def ifthenelse(self, in1, in2, **kwargs):
        for match_image in [in1, in2, self]:
            if isinstance(match_image, pyvips.Image):
                break

        if not isinstance(in1, pyvips.Image):
            in1 = Image._imageize(match_image, in1)
        if not isinstance(in2, pyvips.Image):
            in2 = Image._imageize(match_image, in2)

        return pyvips.Operation.call('ifthenelse', self, in1, in2, **kwargs)

    def scaleimage(self, **kwargs):
        """Scale an image to 0 - 255.

        This is the libvips ``scale`` operation, renamed to avoid a clash with
        the ``scale`` for convolution masks.

        Keyword args:
            exp (float): Exponent for log scale.
            log (bool): Switch to turn on log scaling.

        Returns:
            A new :class:`Image`.

        Raises:
            :class:`.Error`

        """
        return pyvips.Operation.call('scale', self, **kwargs)


__all__ = ['Image']



# vim: set fileencoding=utf-8 :
import pytest

import pyvips
from helpers import JPEG_FILE, assert_almost_equal_objects


class TestGValue:
    def test_bool(self):
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.gbool_type)
        gv.set(True)
        value = gv.get()
        assert value

        gv.set(False)
        value = gv.get()
        assert not value

    def test_int(self):
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.gint_type)
        gv.set(12)
        value = gv.get()
        assert value == 12

    def test_uint64(self):
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.guint64_type)
        gv.set(2 ** 64 - 1)  # G_MAXUINT64
        value = gv.get()
        assert value == 2 ** 64 - 1

    def test_double(self):
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.gdouble_type)
        gv.set(3.1415)
        value = gv.get()
        assert value == 3.1415

    def test_enum(self):
        # the Interpretation enum is created when the first image is made --
        # make it ourselves in case we are run before the first image
        pyvips.vips_lib.vips_interpretation_get_type()
        interpretation_gtype = pyvips.gobject_lib. \
            g_type_from_name(b'VipsInterpretation')
        gv = pyvips.GValue()
        gv.set_type(interpretation_gtype)
        gv.set('xyz')
        value = gv.get()
        assert value == 'xyz'

    def test_flags(self):
        # the OperationFlags enum is created when the first op is made --
        # make it ourselves in case we are run before that
        pyvips.vips_lib.vips_operation_flags_get_type()
        operationflags_gtype = pyvips.gobject_lib. \
            g_type_from_name(b'VipsOperationFlags')
        gv = pyvips.GValue()
        gv.set_type(operationflags_gtype)
        gv.set(12)
        value = gv.get()
        assert value == 12

        # we also support setting flags with strings
        gv.set("deprecated")
        value = gv.get()
        assert value == 8

        # libvips 8.15 allows this as well
        # gv.set("deprecated|nocache")
        # though we don't test it

    def test_string(self):
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.gstr_type)
        gv.set('banana')
        value = gv.get()
        assert value == 'banana'

    def test_array_int(self):
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.array_int_type)
        gv.set([1, 2, 3])
        value = gv.get()
        assert_almost_equal_objects(value, [1, 2, 3])

    def test_array_double(self):
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.array_double_type)
        gv.set([1.1, 2.1, 3.1])
        value = gv.get()
        assert_almost_equal_objects(value, [1.1, 2.1, 3.1])

    def test_image(self):
        image = pyvips.Image.new_from_file(JPEG_FILE)
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.image_type)
        gv.set(image)
        value = gv.get()
        assert value == image

    def test_array_image(self):
        image = pyvips.Image.new_from_file(JPEG_FILE)
        r, g, b = image.bandsplit()
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.array_image_type)
        gv.set([r, g, b])
        value = gv.get()
        assert value, [r, g == b]

    def test_blob(self):
        with open(JPEG_FILE, 'rb') as f:
            blob = f.read()
        gv = pyvips.GValue()
        gv.set_type(pyvips.GValue.blob_type)
        gv.set(blob)
        value = gv.get()
        assert value == blob


if __name__ == '__main__':
    pytest.main()



# wrap VipsRegion

import pyvips
from pyvips import ffi, glib_lib, vips_lib, Error, at_least_libvips


class Region(pyvips.VipsObject):
    """Wrap a VipsRegion object.

    """

    def __init__(self, pointer):
        # logger.debug('Image.__init__: pointer = %s', pointer)
        super(Region, self).__init__(pointer)

    # constructors

    @staticmethod
    def new(image):
        """Make a region on an image.

        Returns:
            A new :class:`.Region`.

        Raises:
            :class:`.Error`

        """

        pointer = vips_lib.vips_region_new(image.pointer)
        if pointer == ffi.NULL:
            raise Error('unable to make region')

        return pyvips.Region(pointer)

    def width(self):
        """Width of pixels held by region."""
        if not at_least_libvips(8, 8):
            raise Error('libvips too old')

        return vips_lib.vips_region_width(self.pointer)

    def height(self):
        """Height of pixels held by region."""
        if not at_least_libvips(8, 8):
            raise Error('libvips too old')

        return vips_lib.vips_region_height(self.pointer)

    def fetch(self, x, y, w, h):
        """Fill a region with pixel data.

        Pixels are filled with data!

        Returns:
            Pixel data.

        Raises:
            :class:`.Error`

        """

        if not at_least_libvips(8, 8):
            raise Error('libvips too old')

        psize = ffi.new('size_t *')
        pointer = vips_lib.vips_region_fetch(self.pointer, x, y, w, h, psize)
        if pointer == ffi.NULL:
            raise Error('unable to fetch from region')

        pointer = ffi.gc(pointer, glib_lib.g_free)
        return ffi.buffer(pointer, psize[0])


__all__ = ['Region']



import logging

import pyvips
from pyvips import ffi, vips_lib, at_least_libvips

logger = logging.getLogger(__name__)


class TargetCustom(pyvips.Target):
    """An output target you can connect action signals to to implement
    behaviour.

    """

    def __init__(self):
        """Make a new target you can customise.

        You can pass this target to (for example) :meth:`write_to_target`.

        """

        target = ffi.cast('VipsTarget*', vips_lib.vips_target_custom_new())
        super(TargetCustom, self).__init__(target)

    def on_write(self, handler):
        """Attach a write handler.

        The interface is exactly as io.write(). The handler is given a
        bytes-like object to write, and should return the number of bytes
        written.

        """

        def interface_handler(buf):
            return handler(buf)

        self.signal_connect("write", interface_handler)

    def on_read(self, handler):
        """Attach a read handler.

        The interface is exactly as io.read(). The handler is given a number
        of bytes to fetch, and should return a bytes-like object containing up
        to that number of bytes. If there is no more data available, it should
        return None.

        Read handlers are optional for targets. If you do not set one, your
        target will be treated as unreadable and libvips will be unable to
        write some file types (just TIFF, as of the time of writing).

        """

        def interface_handler(buf):
            chunk = handler(len(buf))
            if chunk is None:
                return 0

            bytes_read = len(chunk)
            buf[:bytes_read] = chunk

            return bytes_read

        if at_least_libvips(8, 13):
            self.signal_connect("read", interface_handler)

    def on_seek(self, handler):
        """Attach a seek handler.

        The interface is the same as io.seek(), so the handler is passed
        parameters for offset and whence with the same meanings.

        However, the handler MUST return the new seek position. A simple way
        to do this is to call io.tell() and return that result.

        Seek handlers are optional. If you do not set one, your target will be
        treated as unseekable and libvips will be unable to write some file
        types (just TIFF, as of the time of writing).

        """

        if at_least_libvips(8, 13):
            self.signal_connect("seek", handler)

    def on_end(self, handler):
        """Attach an end handler.

        This optional handler is called at the end of write. It should do any
        cleaning up necessary, and return 0 on success and -1 on error.

        """

        if not at_least_libvips(8, 13):
            # fall back for older libvips
            self.on_finish(handler)
        else:
            self.signal_connect("end", handler)

    def on_finish(self, handler):
        """Attach a finish handler.

        For libvips 8.13 and later, this method is deprecated in favour of
        :meth:`on_end`.

        """

        self.signal_connect("finish", handler)


__all__ = ['TargetCustom']



import logging

import pyvips
from pyvips import ffi, gobject_lib, _to_bytes, Error, type_name, \
    at_least_libvips

logger = logging.getLogger(__name__)

# the python marshalers for gobject signal handling
# - we keep a ref to each callback to stop them being GCd
# - I tried to make this less copy-paste, but failed -- check again

if pyvips.API_mode:
    @ffi.def_extern()
    def _marshal_image_progress(vi, pointer, handle):
        gobject_lib.g_object_ref(vi)
        image = pyvips.Image(vi)
        callback = ffi.from_handle(handle)
        progress = ffi.cast('VipsProgress*', pointer)
        callback(image, progress)
    _marshal_image_progress_cb =  \
        ffi.cast('GCallback', gobject_lib._marshal_image_progress)
else:
    @ffi.callback('void(VipsImage*, void*, void*)')
    def _marshal_image_progress(vi, pointer, handle):
        gobject_lib.g_object_ref(vi)
        image = pyvips.Image(vi)
        callback = ffi.from_handle(handle)
        progress = ffi.cast('VipsProgress*', pointer)
        callback(image, progress)
    _marshal_image_progress_cb = \
        ffi.cast('GCallback', _marshal_image_progress)

_marshalers = {
    'preeval': _marshal_image_progress_cb,
    'eval': _marshal_image_progress_cb,
    'posteval': _marshal_image_progress_cb,
}

if at_least_libvips(8, 9):
    if pyvips.API_mode:
        @ffi.def_extern()
        def _marshal_read(gobject, pointer, length, handle):
            buf = ffi.buffer(pointer, length)
            callback = ffi.from_handle(handle)
            return callback(buf)
        _marshal_read_cb = ffi.cast('GCallback', gobject_lib._marshal_read)
    else:
        @ffi.callback('gint64(VipsSourceCustom*, void*, gint64, void*)')
        def _marshal_read(gobject, pointer, length, handle):
            buf = ffi.buffer(pointer, length)
            callback = ffi.from_handle(handle)
            return callback(buf)
        _marshal_read_cb = ffi.cast('GCallback', _marshal_read)
    _marshalers['read'] = _marshal_read_cb

    if pyvips.API_mode:
        @ffi.def_extern()
        def _marshal_seek(gobject, offset, whence, handle):
            callback = ffi.from_handle(handle)
            return callback(offset, whence)
        _marshal_seek_cb = \
            ffi.cast('GCallback', gobject_lib._marshal_seek)
    else:
        @ffi.callback('gint64(VipsSourceCustom*, gint64, int, void*)')
        def _marshal_seek(gobject, offset, whence, handle):
            callback = ffi.from_handle(handle)
            return callback(offset, whence)
        _marshal_seek_cb = ffi.cast('GCallback', _marshal_seek)
    _marshalers['seek'] = _marshal_seek_cb

    if pyvips.API_mode:
        @ffi.def_extern()
        def _marshal_write(gobject, pointer, length, handle):
            buf = ffi.buffer(pointer, length)
            callback = ffi.from_handle(handle)
            return callback(buf)
        _marshal_write_cb = ffi.cast('GCallback', gobject_lib._marshal_write)
    else:
        @ffi.callback('gint64(VipsTargetCustom*, void*, gint64, void*)')
        def _marshal_write(gobject, pointer, length, handle):
            buf = ffi.buffer(pointer, length)
            callback = ffi.from_handle(handle)
            return callback(buf)
        _marshal_write_cb = ffi.cast('GCallback', _marshal_write)
    _marshalers['write'] = _marshal_write_cb

    if pyvips.API_mode:
        @ffi.def_extern()
        def _marshal_finish(gobject, handle):
            callback = ffi.from_handle(handle)
            callback()
        _marshal_finish_cb = ffi.cast('GCallback', gobject_lib._marshal_finish)
    else:
        @ffi.callback('void(VipsTargetCustom*, void*)')
        def _marshal_finish(gobject, handle):
            callback = ffi.from_handle(handle)
            callback()
        _marshal_finish_cb = ffi.cast('GCallback', _marshal_finish)
    _marshalers['finish'] = _marshal_finish_cb

if at_least_libvips(8, 13):
    if pyvips.API_mode:
        @ffi.def_extern()
        def _marshal_end(gobject, handle):
            callback = ffi.from_handle(handle)
            return callback()
        _marshal_end_cb = ffi.cast('GCallback', gobject_lib._marshal_end)
    else:
        @ffi.callback('int(VipsTargetCustom*, void*)')
        def _marshal_end(gobject, handle):
            callback = ffi.from_handle(handle)
            return callback()
        _marshal_end_cb = ffi.cast('GCallback', _marshal_end)
    _marshalers['end'] = _marshal_end_cb


class GObject(object):
    """Manage GObject lifetime.

    """
    __slots__ = ('_handles', 'pointer')

    def __init__(self, pointer):
        """Wrap around a pointer.

        Wraps a GObject instance around an underlying pointer. When the
        instance is garbage-collected, the underlying object is unreferenced.

        """

        # we have to record all of the ffi.new_handle we make for callbacks on
        # this object to prevent them being GC'd
        self._handles = []

        # record the pointer we were given to manage
        # on GC, unref
        self.pointer = ffi.gc(pointer, gobject_lib.g_object_unref)
        # logger.debug('GObject.__init__: pointer = %s', str(self.pointer))

    @staticmethod
    def new_pointer_from_gtype(gtype):
        """Make a new GObject pointer from a gtype.

        This is useful for subclasses which need to control the construction
        process.

        You can pass the result pointer to the Python constructor for the
        object you are building. You will need to call VipsObject.build() to
        finish construction.

        Returns:
            A pointer to a new GObject.

        Raises:
            :class:`.Error`

        """

        pointer = gobject_lib.g_object_new(gtype, ffi.NULL)
        if pointer == ffi.NULL:
            raise Error(f"can't create {type_name(gtype)}")

        return pointer

    def signal_connect(self, name, callback):
        """Connect to a signal on this object.

        The callback will be triggered every time this signal is issued on this
        instance. It will be passed the image ('self' here), and a single
        `void *` pointer from libvips.

        The value of the pointer, if any, depends on the signal -- for
        example, ::eval passes a pointer to a `VipsProgress` struct.

        """

        if name not in _marshalers:
            raise Error(f'unsupported signal "{name}"')

        go = ffi.cast('GObject *', self.pointer)
        handle = ffi.new_handle(callback)
        # we need to keep refs to the ffi handle and the callback to prevent
        # them being GCed
        # the callback might be a bound method (a closure) rather than a simple
        # function, so it can vanish
        self._handles.append(handle)
        self._handles.append(callback)

        gobject_lib.g_signal_connect_data(go, _to_bytes(name),
                                          _marshalers[name],
                                          handle, ffi.NULL, 0)


__all__ = ['GObject']



import logging

import pyvips
from pyvips import ffi, vips_lib

logger = logging.getLogger(__name__)


class SourceCustom(pyvips.Source):
    """Custom sources allow reading data from otherwise unsupported sources.
    Requires libvips `>= 8.9.0`.

    To use, create a SourceCustom object, then provide callbacks to
    :meth:`on_read` and :meth:`on_seek`.
    """

    def __init__(self):
        """Make a new custom source.

        You can pass this source to (for example) :meth:`new_from_source`.

        """

        source = ffi.cast('VipsSource*', vips_lib.vips_source_custom_new())
        super(SourceCustom, self).__init__(source)

    def on_read(self, handler):
        """Attach a read handler.

        The interface is exactly as io.read(). The handler is given a number
        of bytes to fetch, and should return a bytes-like object containing up
        to that number of bytes. If there is no more data available, it should
        return None.

        """

        def interface_handler(buf):
            chunk = handler(len(buf))
            if chunk is None:
                return 0

            bytes_read = len(chunk)
            buf[:bytes_read] = chunk

            return bytes_read

        self.signal_connect("read", interface_handler)

    def on_seek(self, handler):
        """Attach a seek handler.

        The interface is the same as io.seek(), so the handler is passed
        parameters for offset and whence with the same meanings.

        However, the handler MUST return the new seek position. A simple way
        to do this is to call io.tell() and return that result.

        Seek handlers are optional. If you do not set one, your source will be
        treated as unseekable and libvips will do extra caching.

        """

        self.signal_connect("seek", handler)


__all__ = ['SourceCustom']



# vim: set fileencoding=utf-8 :

import pytest

import pyvips


class TestProgress:
    def test_progress(self):
        seen_preeval = False
        seen_eval = False
        seen_posteval = False

        def preeval_cb(image, progress):
            nonlocal seen_preeval
            seen_preeval = True

        def eval_cb(image, progress):
            nonlocal seen_eval
            seen_eval = True

        def posteval_cb(image, progress):
            nonlocal seen_posteval
            seen_posteval = True

        image = pyvips.Image.black(1, 100000)
        image.set_progress(True)
        image.signal_connect('preeval', preeval_cb)
        image.signal_connect('eval', eval_cb)
        image.signal_connect('posteval', posteval_cb)
        image.avg()

        assert seen_preeval
        assert seen_eval
        assert seen_posteval

    def test_progress_fields(self):
        def preeval_cb(image, progress):
            assert progress.run == 0
            assert progress.eta == 0
            assert progress.percent == 0
            assert progress.tpels == 10000
            assert progress.npels == 0

        def eval_cb(image, progress):
            pass

        def posteval_cb(image, progress):
            assert progress.percent == 100
            assert progress.tpels == 10000
            assert progress.npels == 10000

        image = pyvips.Image.black(10, 1000)
        image.set_progress(True)
        image.signal_connect('preeval', preeval_cb)
        image.signal_connect('eval', eval_cb)
        image.signal_connect('posteval', posteval_cb)
        image.avg()

    def test_progress_kill(self):
        def preeval_cb(image, progress):
            pass

        def eval_cb(image, progress):
            image.set_kill(True)

        def posteval_cb(image, progress):
            pass

        # has to be very tall to ensure the kill has enough threadpool loops
        # to work
        image = pyvips.Image.black(1, 1000000)
        image.set_progress(True)
        image.signal_connect('preeval', preeval_cb)
        image.signal_connect('eval', eval_cb)
        image.signal_connect('posteval', posteval_cb)

        with pytest.raises(Exception):
            image.copy_memory()



import logging
import numbers

import pyvips
from pyvips import ffi, vips_lib, gobject_lib, \
    glib_lib, Error, _to_bytes, _to_string, type_name, type_from_name, \
    at_least_libvips

logger = logging.getLogger(__name__)


class GValue(object):

    """Wrap GValue in a Python class.

    This class wraps :class:`.GValue` in a convenient interface. You can use
    instances of this class to get and set :class:`.GObject` properties.

    On construction, :class:`.GValue` is all zero (empty). You can pass it to
    a get function to have it filled by :class:`.GObject`, or use init to
    set a type, set to set a value, then use it to set an object property.

    GValue lifetime is managed automatically.

    """
    __slots__ = ('pointer', 'gvalue')

    # look up some common gtypes at init for speed
    gbool_type = type_from_name('gboolean')
    gint_type = type_from_name('gint')
    guint64_type = type_from_name('guint64')
    gdouble_type = type_from_name('gdouble')
    gstr_type = type_from_name('gchararray')
    genum_type = type_from_name('GEnum')
    gflags_type = type_from_name('GFlags')
    gobject_type = type_from_name('GObject')
    image_type = type_from_name('VipsImage')
    array_int_type = type_from_name('VipsArrayInt')
    array_double_type = type_from_name('VipsArrayDouble')
    array_image_type = type_from_name('VipsArrayImage')
    refstr_type = type_from_name('VipsRefString')
    blob_type = type_from_name('VipsBlob')
    source_type = type_from_name('VipsSource')
    target_type = type_from_name('VipsTarget')

    pyvips.vips_lib.vips_band_format_get_type()
    format_type = type_from_name('VipsBandFormat')

    if at_least_libvips(8, 6):
        pyvips.vips_lib.vips_blend_mode_get_type()
    blend_mode_type = type_from_name('VipsBlendMode')

    # map a gtype to the name of the corresponding Python type
    _gtype_to_python = {
        gbool_type: 'bool',
        gint_type: 'int',
        guint64_type: 'long',  # Note: int and long are unified in Python 3
        gdouble_type: 'float',
        gstr_type: 'str',
        refstr_type: 'str',
        genum_type: 'str',
        gflags_type: 'int',
        gobject_type: 'GObject',
        image_type: 'Image',
        array_int_type: 'list[int]',
        array_double_type: 'list[float]',
        array_image_type: 'list[Image]',
        blob_type: 'str',
        source_type: 'Source',
        target_type: 'Target',
    }

    @staticmethod
    def gtype_to_python(gtype):
        """Map a gtype to the name of the Python type we use to represent it.

        """

        fundamental = gobject_lib.g_type_fundamental(gtype)

        # enums can be strings or class members ... we want to generate a union
        # type
        if fundamental == GValue.genum_type:
            name = type_name(gtype)
            if name.startswith('Vips'):
                name = name[4:]
            return f"Union[str, {name}]"
        if gtype in GValue._gtype_to_python:
            return GValue._gtype_to_python[gtype]
        if fundamental in GValue._gtype_to_python:
            return GValue._gtype_to_python[fundamental]
        return '<unknown type>'

    @staticmethod
    def to_enum(gtype, value):
        """Turn a string into an enum value ready to be passed into libvips.

        """

        if isinstance(value, str):
            enum_value = vips_lib.vips_enum_from_nick(b'pyvips', gtype,
                                                      _to_bytes(value))
            if enum_value < 0:
                raise Error('no value {0} in gtype {1} ({2})'.
                            format(value, type_name(gtype), gtype))
        else:
            enum_value = value

        return enum_value

    @staticmethod
    def from_enum(gtype, enum_value):
        """Turn an int back into an enum string.

        """

        pointer = vips_lib.vips_enum_nick(gtype, enum_value)
        if pointer == ffi.NULL:
            raise Error('value not in enum')

        return _to_string(pointer)

    @staticmethod
    def to_flag(gtype, value):
        """Turn a string into a flag value ready to be passed into libvips.

        """

        if isinstance(value, str):
            flag_value = vips_lib.vips_flags_from_nick(b'pyvips', gtype,
                                                       _to_bytes(value))
            if flag_value < 0:
                raise Error('no value {0} in gtype {1} ({2})'.
                            format(value, type_name(gtype), gtype))
        else:
            flag_value = value

        return flag_value

    def __init__(self):
        # allocate memory for the gvalue which will be freed on GC
        self.pointer = ffi.new('GValue *')
        # logger.debug('GValue.__init__: pointer = %s', self.pointer)

        # and tag it to be unset on GC as well
        self.gvalue = ffi.gc(self.pointer, gobject_lib.g_value_unset)
        # logger.debug('GValue.__init__: gvalue = %s', self.gvalue)

    def set_type(self, gtype):
        """Set the type of a GValue.

        GValues have a set type, fixed at creation time. Use set_type to set
        the type of a GValue before assigning to it.

        GTypes are 32 or 64-bit integers (depending on the platform). See
        type_find.

        """

        gobject_lib.g_value_init(self.gvalue, gtype)

    def set(self, value):
        """Set a GValue.

        The value is converted to the type of the GValue, if possible, and
        assigned.

        """

        # logger.debug('GValue.set: value = %s', value)

        gtype = self.gvalue.g_type
        fundamental = gobject_lib.g_type_fundamental(gtype)

        if gtype == GValue.gbool_type:
            gobject_lib.g_value_set_boolean(self.gvalue, value)
        elif gtype == GValue.gint_type:
            gobject_lib.g_value_set_int(self.gvalue, int(value))
        elif gtype == GValue.guint64_type:
            gobject_lib.g_value_set_uint64(self.gvalue, value)
        elif gtype == GValue.gdouble_type:
            gobject_lib.g_value_set_double(self.gvalue, value)
        elif fundamental == GValue.genum_type:
            gobject_lib.g_value_set_enum(self.gvalue,
                                         GValue.to_enum(gtype, value))
        elif fundamental == GValue.gflags_type:
            gobject_lib.g_value_set_flags(self.gvalue,
                                          GValue.to_flag(gtype, value))
        elif gtype == GValue.gstr_type:
            gobject_lib.g_value_set_string(self.gvalue, _to_bytes(value))
        elif gtype == GValue.refstr_type:
            vips_lib.vips_value_set_ref_string(self.gvalue, _to_bytes(value))
        elif fundamental == GValue.gobject_type:
            gobject_lib.g_value_set_object(self.gvalue, value.pointer)
        elif gtype == GValue.array_int_type:
            if isinstance(value, numbers.Number):
                value = [value]

            array = ffi.new('int[]', value)
            vips_lib.vips_value_set_array_int(self.gvalue, array, len(value))
        elif gtype == GValue.array_double_type:
            if isinstance(value, numbers.Number):
                value = [value]

            array = ffi.new('double[]', value)
            vips_lib.vips_value_set_array_double(self.gvalue, array,
                                                 len(value))
        elif gtype == GValue.array_image_type:
            if isinstance(value, pyvips.Image):
                value = [value]

            vips_lib.vips_value_set_array_image(self.gvalue, len(value))
            array = vips_lib.vips_value_get_array_image(self.gvalue, ffi.NULL)
            for i, image in enumerate(value):
                gobject_lib.g_object_ref(image.pointer)
                array[i] = image.pointer
        elif gtype == GValue.blob_type:
            # we need to set the blob to a copy of the string that vips_lib
            # can own
            memory = glib_lib.g_malloc(len(value))
            ffi.memmove(memory, value, len(value))

            # In API mode, we use set_blob_free in a backwards compatible way.
            # For pre-8.6 libvipses, in ABI mode, we declare the type of the
            # free func in set_blob incorrectly so that we can pass g_free
            # at runtime without triggering an exception.
            if pyvips.API_mode or at_least_libvips(8, 6):
                vips_lib.vips_value_set_blob_free(self.gvalue,
                                                  memory, len(value))
            else:
                vips_lib.vips_value_set_blob(self.gvalue,
                                             glib_lib.g_free,
                                             memory, len(value))
        else:
            raise Error('unsupported gtype for set {0}, fundamental {1}'.
                        format(type_name(gtype), type_name(fundamental)))

    def get(self):
        """Get the contents of a GValue.

        The contents of the GValue are read out as a Python type.
        """

        # logger.debug('GValue.get: self = %s', self)

        gtype = self.gvalue.g_type
        fundamental = gobject_lib.g_type_fundamental(gtype)

        result = None

        if gtype == GValue.gbool_type:
            result = bool(gobject_lib.g_value_get_boolean(self.gvalue))
        elif gtype == GValue.gint_type:
            result = gobject_lib.g_value_get_int(self.gvalue)
        elif gtype == GValue.guint64_type:
            result = gobject_lib.g_value_get_uint64(self.gvalue)
        elif gtype == GValue.gdouble_type:
            result = gobject_lib.g_value_get_double(self.gvalue)
        elif fundamental == GValue.genum_type:
            return GValue.from_enum(gtype,
                                    gobject_lib.g_value_get_enum(self.gvalue))
        elif fundamental == GValue.gflags_type:
            result = gobject_lib.g_value_get_flags(self.gvalue)
        elif gtype == GValue.gstr_type:
            pointer = gobject_lib.g_value_get_string(self.gvalue)

            if pointer != ffi.NULL:
                result = _to_string(pointer)
        elif gtype == GValue.refstr_type:
            psize = ffi.new('size_t *')
            pointer = vips_lib.vips_value_get_ref_string(self.gvalue, psize)

            # psize[0] will be number of bytes in string, but just assume it's
            # NULL-terminated
            result = _to_string(pointer)
        elif gtype == GValue.image_type:
            # g_value_get_object() will not add a ref ... that is
            # held by the gvalue
            go = gobject_lib.g_value_get_object(self.gvalue)
            vi = ffi.cast('VipsImage *', go)

            # we want a ref that will last with the life of the vimage:
            # this ref is matched by the unref that's attached to finalize
            # by Image()
            gobject_lib.g_object_ref(vi)

            result = pyvips.Image(vi)
        elif gtype == GValue.array_int_type:
            pint = ffi.new('int *')
            array = vips_lib.vips_value_get_array_int(self.gvalue, pint)

            result = []
            for i in range(0, pint[0]):
                result.append(array[i])
        elif gtype == GValue.array_double_type:
            pint = ffi.new('int *')
            array = vips_lib.vips_value_get_array_double(self.gvalue, pint)

            result = []
            for i in range(0, pint[0]):
                result.append(array[i])
        elif gtype == GValue.array_image_type:
            pint = ffi.new('int *')
            array = vips_lib.vips_value_get_array_image(self.gvalue, pint)

            result = []
            for i in range(0, pint[0]):
                vi = array[i]
                gobject_lib.g_object_ref(vi)
                image = pyvips.Image(vi)
                result.append(image)
        elif gtype == GValue.blob_type:
            psize = ffi.new('size_t *')
            array = vips_lib.vips_value_get_blob(self.gvalue, psize)
            buf = ffi.cast('char*', array)

            result = ffi.unpack(buf, psize[0])
        else:
            raise Error('unsupported gtype for get {0}'.
                        format(type_name(gtype)))

        return result


__all__ = ['GValue']



# basic defs and link to ffi


from pyvips import ffi, glib_lib, vips_lib, gobject_lib, \
    _to_string, _to_bytes, Error


def leak_set(leak):
    """Enable or disable libvips leak checking.

    With this enabled, libvips will check for object and area leaks on exit.
    Enabling this option will make libvips run slightly more slowly.
    """

    return vips_lib.vips_leak_set(leak)


def shutdown():
    """Shut libvips down."""
    vips_lib.vips_shutdown()


def version(flag):
    """Get the major, minor or micro version number of the libvips library.

    Args:
        flag (int): Pass flag 0 to get the major version number, flag 1 to
            get minor, flag 2 to get micro.

    Returns:
        The version number,

    Raises:
        :class:`.Error`
    """

    value = vips_lib.vips_version(flag)
    if value < 0:
        raise Error('unable to get library version')

    return value


def get_suffixes():
    """Get a list of all the filename suffixes supported by libvips.

    Returns:
        [string]

    """

    names = []

    if at_least_libvips(8, 8):
        array = vips_lib.vips_foreign_get_suffixes()
        i = 0
        while array[i] != ffi.NULL:
            name = _to_string(array[i])
            if name not in names:
                names.append(name)
            glib_lib.g_free(array[i])
            i += 1
        glib_lib.g_free(array)

    return names


# we need to define this before we import the declarations: they need to know
# which bits to make
def at_least_libvips(x, y):
    """Is this at least libvips x.y?"""

    major = version(0)
    minor = version(1)

    return major > x or (major == x and minor >= y)


def type_find(basename, nickname):
    """Get the GType for a name.

    Looks up the GType for a nickname. Types below basename in the type
    hierarchy are searched.
    """

    return vips_lib.vips_type_find(_to_bytes(basename), _to_bytes(nickname))


def type_name(gtype):
    """Return the name for a GType."""

    return _to_string(gobject_lib.g_type_name(gtype))


def nickname_find(gtype):
    """Return the nickname for a GType."""

    return _to_string(vips_lib.vips_nickname_find(gtype))


def type_from_name(name):
    """Return the GType for a name."""

    return gobject_lib.g_type_from_name(_to_bytes(name))


def type_map(gtype, fn):
    """Map fn over all child types of gtype."""

    cb = ffi.callback('VipsTypeMap2Fn', fn)
    return vips_lib.vips_type_map(gtype, cb, ffi.NULL, ffi.NULL)


def values_for_enum(gtype):
    """Deprecated."""

    g_type_class = gobject_lib.g_type_class_ref(gtype)
    g_enum_class = ffi.cast('GEnumClass *', g_type_class)

    result = [_to_string(g_enum_class.values[i].value_nick)
              for i in range(g_enum_class.n_values)]
    # FIXME: remove after https://github.com/libvips/libvips/pull/4520
    if 'last' in result:
        result.remove('last')
    return result


def values_for_flag(gtype):
    """Deprecated."""

    g_type_class = gobject_lib.g_type_class_ref(gtype)
    g_flags_class = ffi.cast('GFlagsClass *', g_type_class)

    return [_to_string(g_flags_class.values[i].value_nick)
            for i in range(g_flags_class.n_values)]


def enum_dict(gtype):
    """Get name -> value dict for a enum (gtype)."""

    g_type_class = gobject_lib.g_type_class_ref(gtype)
    g_enum_class = ffi.cast('GEnumClass *', g_type_class)

    result = {_to_string(g_enum_class.values[i].value_nick):
              g_enum_class.values[i].value
              for i in range(g_enum_class.n_values)}
    # FIXME: remove after https://github.com/libvips/libvips/pull/4520
    result.pop('last', None)
    return result


def flags_dict(gtype):
    """Get name -> value dict for a flags (gtype)."""

    g_type_class = gobject_lib.g_type_class_ref(gtype)
    g_flags_class = ffi.cast('GFlagsClass *', g_type_class)

    return {_to_string(g_flags_class.values[i].value_nick):
            g_flags_class.values[i].value
            for i in range(g_flags_class.n_values)}


__all__ = [
    'leak_set',
    'shutdown',
    'version',
    'at_least_libvips',
    'type_find',
    'nickname_find',
    'get_suffixes',
    'type_name',
    'type_map',
    'type_from_name',
    'values_for_enum',
    'values_for_flag',
    'enum_dict',
    'flags_dict'
]
